//==============================================================================
// rexpx6x70d_drv.c
//   REX-PE60D/64D/70D, REX-PCI60D/64D/70D DIOドライバ
//   ※DIPSWで識別される4枚分までのサポートになります。
//     DIPSWの設定が重複している場合、後で検出した方は動作しません。
//
//  Copyright (C) 2012 RATOC Systems, Inc.
//  Dual licensed under the BSD (BSD-LICENSE.txt)
//  and GPL (GPL-LICENSE.txt) licenses.
//
//==============================================================================
//
// PCI Express
// REX-PE60D  2port RS-232C     XR17V352 VID=13A8, DID=0352, SSID=6005, SVID=1195(RATOC)
// REX-PE70D  2port RS-422/485  XR17V352 VID=13A8, DID=0352, SSID=6006, SVID=1195(RATOC)
// REX-PE64D  4port RS-232C     XR17V354 VID=13A8, DID=0354, SSID=600B, SVID=1195(RATOC)
//
// PCI
// REX-PCI60D 2port RS-232C     XR17D152 VID=13A8, DID=0152, SSID=6007, SVID=1195(RATOC)
// REX-PCI70D 2port RS-422/485  XR17D152 VID=13A8, DID=0152, SSID=6008, SVID=1195(RATOC)
// REX-PCI64D 4port RS-232C     XR17D154 VID=13A8, DID=0154, SSID=600C, SVID=1195(RATOC)
//
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/ioctl.h>
#include <linux/irq.h>
#include <linux/clk.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/io.h>
#include <linux/semaphore.h>
#include <linux/version.h>
#include <linux/timer.h>

#include "rexpx6x70d_drv.h"

//==============================================================================
// 各種定義
//==============================================================================

//------------------------------------------------------------------------------
// デバッグ用の定義
//------------------------------------------------------------------------------
//#define USE_DEBUG_PROC_REXDIO	// 定義すると /proc/rexdio が有効になる(但し中身は未実装)
//#define USE_DEBUG_IRQ_HANDLER	// 定義すると割り込み要因をチェックせずにMPIOINT割り込みとして処理する
#if 0
#define DEBUG_PRINTK			printk
#else
#define DEBUG_PRINTK(fmt...)	do { } while(0)
#endif

//------------------------------------------------------------------------------
// バージョン番号
//------------------------------------------------------------------------------
#define DRIVER_VERSION			"1.01"

//------------------------------------------------------------------------------
// TRUE/FALSE をマクロ定義
//------------------------------------------------------------------------------
#ifndef TRUE
# define TRUE	(1==1)
#endif
#ifndef FALSE
# define FALSE	(1==0)
#endif

//------------------------------------------------------------------------------
// BOOL の型定義
//------------------------------------------------------------------------------
#ifndef BOOL
# define BOOL	int
#endif

//------------------------------------------------------------------------------
// BYTE/WORD/DWORD の型定義
//------------------------------------------------------------------------------
typedef uint8_t		BYTE;
typedef uint16_t	WORD;
typedef uint32_t	DWORD;

//------------------------------------------------------------------------------
// デバイスID構造体定義
//------------------------------------------------------------------------------
#define VENDOR_ID_EXAR			0x13A8		// EXAR のベンダID

#define VENDOR_IO_RATOC			0x1195		// RATOC のベンダID

#define DEVICE_ID_XR17V352		0x0352		// XR17V352 のデバイスID
#define DEVICE_ID_XR17V354		0x0354		// XR17V354 のデバイスID
#define DEVICE_ID_XR17V152		0x0152		// XR17V152 のデバイスID
#define DEVICE_ID_XR17V154		0x0154		// XR17V154 のデバイスID

#define SUBSYS_ID_REX_PE60D		0x6005		// REX-PE60D のサブシステムID
#define SUBSYS_ID_REX_PE70D		0x6006		// REX-PE70D のサブシステムID
#define SUBSYS_ID_REX_PCI60D	0x6007		// REX-PCI60D のサブシステムID
#define SUBSYS_ID_REX_PCI70D	0x6008		// REX-PCI70D のサブシステムID
#define SUBSYS_ID_REX_PE64D		0x600B		// REX-PE64D のサブシステムID
#define SUBSYS_ID_REX_PCI64D	0x600C		// REX-PCI64D のサブシステムID


//------------------------------------------------------------------------------
// ボードの同時使用最大枚数
//------------------------------------------------------------------------------
#define REX_DIO_MAX_BOARDS		4		// ID=0～3の計4枚まで

//------------------------------------------------------------------------------
// ドライバ/デバイス名称定義
//------------------------------------------------------------------------------
#if defined(_USE_XR17C15X) && defined(_USE_XR17V35X)
#define DRIVER_NAME			"rexPx6x70D"	// ドライバ名称
#define CLASS_DEVICE_NAME	"rexPx6x70D"	// クラスデバイス名称
#define CHR_DEVICE_NAME		"rexdio"		// キャラクタデバイス名称(/dev/recdio<n>)
#define PROC_NAME			"rexdio"		// デバッグ用 proc デバイス名称(/prox/rexdio)
#define VER_MSG				"REX-PCIxxD/PExxD"	// バージョンメッセージ
#elif defined(_USE_XR17C15X)
#define DRIVER_NAME			"rexPx6x70Dpci"	// ドライバ名称
#define CLASS_DEVICE_NAME	"rexPx6x70Dpci"	// クラスデバイス名称
#define CHR_DEVICE_NAME		"rexdioPCI"		// キャラクタデバイス名称(/dev/recdioPCI<n>)
#define PROC_NAME			"rexdioPCI"		// デバッグ用 proc デバイス名称(/prox/rexdioPCI)
#define VER_MSG				"REX-PCI6x/70D"	// バージョンメッセージ
#elif defined(_USE_XR17V35X)
#define DRIVER_NAME			"rexPx6x70Dpe"	// ドライバ名称
#define CLASS_DEVICE_NAME	"rexPx6x70Dpe"	// クラスデバイス名称
#define CHR_DEVICE_NAME		"rexdioPE"		// キャラクタデバイス名称(/dev/recdioPCI<n>)
#define PROC_NAME			"rexdioPE"		// デバッグ用 proc デバイス名称(/prox/rexdioPCI)
#define VER_MSG				"REX-PE6x/70D"	// バージョンメッセージ
#else
#error not defined _USE_XR17C15X or _USE_XR17V35X.
#endif

//------------------------------------------------------------------------------
// クラスデバイス名
// ※DIPSWで設定するIDに対応する
//------------------------------------------------------------------------------
#if 0
static const char	*s_szClassDeviceName[REX_DIO_MAX_BOARDS] = {
	  CLASS_DEVICE_NAME ".0"		// DIO #0 ⇒ rexPx6x70D.0
	, CLASS_DEVICE_NAME ".1"		// DIO #1 ⇒ rexPx6x70D.1
	, CLASS_DEVICE_NAME ".2"		// DIO #2 ⇒ rexPx6x70D.2
	, CLASS_DEVICE_NAME ".3"		// DIO #3 ⇒ rexPx6x70D.3
};
#else
static const char	*s_szClassDeviceName[REX_DIO_MAX_BOARDS] = {
	  CHR_DEVICE_NAME "0"			// DIO #0 ⇒ rexdio0
	, CHR_DEVICE_NAME "1"			// DIO #1 ⇒ rexdio1
	, CHR_DEVICE_NAME "2"			// DIO #2 ⇒ rexdio2
	, CHR_DEVICE_NAME "3"			// DIO #3 ⇒ rexdio3
};
#endif

//------------------------------------------------------------------------------
// ドライバメジャー番号
//------------------------------------------------------------------------------
#define REX_DIO_MAJOR_NUMBER	0		// 自動割り当て
//#define REX_DIO_MAJOR_NUMBER	500		// 固定割り当て


//==============================================================================
// ハードウェアレジスタ構造体
//==============================================================================

//------------------------------------------------------------------------------
// XR17D154/152 DIO ハードウェアレジスタ構造体
//------------------------------------------------------------------------------
//
// Addr   Register      Read/Write Comment
// ------ ------------- --------------------------------------------------------
// +0080: INT0[7:0]     RO, Interrupt[3:0], Reserved[7:4]
// +0081: INT1[15:8]    RO, Interrupt Source[7:0]
// +0082: INT2[23:16]   RO, Interrupt Source[3:0], Reserved[7:4]
// +0083: INT3[31:24]   <<RESERVED>>
// ------ ------------- --------------------------------------------------------
// +0084: TIMERCNTL     RW, Timer Control
// +0085: TIMER         <<RESERVED>>
// +0086: TIMERLSB      RW, Timer LSB [7:0]
// +0087: TIMERMSB      RW, Timer MSB [15:8]
// ------ ------------- --------------------------------------------------------
// +0088: 8XMODE        RW, UART channel 8X Clock mode enable[3:0]
// +0089: REGA          <<RESERVED>>
// +008A: RESET         WO, Self clear bits after executing Reset[3:0]
// +008B: SLEEP         RW, Sleep mode[3:0]
// ------ ------------- --------------------------------------------------------
// +008C: DREV          RO, Device revision
// +008D: DVID          RO, Device identification
// +008E: REGB          RW, configuration, EEPROM control
// ------ ------------- --------------------------------------------------------
// +008F: MPIOINT       RW, MPIO interrupt mask
// +0090: MPIOLVL       RW, MPIO level control
// +0091: MPIO3T[       RW, MPIO output control
// +0092: MPIOINV       RW, MPIO input polarity select
// +0093: MPIOSEL       RW, MPIO select
// ------ ------------- --------------------------------------------------------
//
// SIO Interrupt (INT0)
//   [0]   CH1
//   [1]   CH2
//   [2]   CH3 (XR17D154)
//   [3]   CH4 (XR17D154)
//   [7:4] <<RESERVED>>
//
// SIO Interrupt source (INT2[3:0]+INT1[7:0])
//   [ 2: 0](INT1[2:0])         CH1 Interrupt Source Encoding
//   [ 5: 3](INT1[5:3])         CH2 Interrupt Source Encoding
//   [ 8: 6](INT2[0]+INT1[7:6]) CH3 Interrupt Source Encoding (XR17D154)
//   [11: 9](INT2[3:1])         CH4 Interrupt Source Encoding (XR17D154)
//   [15:12](INT2[7:4])         <<RESERVED>>
//
// Interrupt Encoding(0～7)
//   0: None
//   1: RXRDY and RX line Status (logic OR of LSR[4:1])
//   2: RXRDY Time-out
//   3: TXRDY, THR or TSR(auto RS485 mode) empty
//   4: MSR, RTS/CTS or DTR/DSR delta or Xoff/Xon det. or special char. detected.
//   5: <<Reserved>>
//   6: MPIO pin(s), Available only within channel 0, reserved in other channels.
//   7: TIMER time-out. Available only within channel 0, reserved in other channels.
//
//   RXRDY is cleared by reading data in th RX FIFO until it falls below the trigger level.
//   RXRDY Time-out is cleared by reading data until the RX FIFO is empty.
//   RX Line Status interrupt clears after reading the LSR register.
//   TXRDY interrupt clears after reading ISR register that is in the UART channel register set.
//   Modem Status Register interrupt clears after reading MSR register that is in the UART channel register set.
//   RTS/CTS or DTR/DSR delta interrupt clears after reading MSR register that is in the UART channel register set.
//   Xoff/Xon interrupt clears after reading the ISR register that is in the UART channel register set.
//   Special character detect interrupt is cleared by read to ISR or after the next character is received.
//   TIMER Time-out interrupt clears after reading the TIMERCNTL register that is in the Device Configuration register set.
//   MPIO interrupt clears after reading MPIOLVL register that is the Device Configuration register set.
//
// Timer Control (TIMERCNTL)
//   [0]   INT Enable
//   [1]   Start/Stop
//   [2]   Single/Re-trigger
//   [3]   Clock Select
//   [4]   MPIO[0] Control
//   [7:5] <<RESERVED>>
//
// 8X Clock Enable (8XMODE)
//   [0]   CH1
//   [1]   CH2
//   [2]   CH3 (XR17D154)
//   [3]   CH4 (XR17D154)
//   [7:4] <<RESERVED>>
//
// UART Reset (RESET)
//   [0]   CH1
//   [1]   CH2
//   [2]   CH3 (XR17D154)
//   [3]   CH4 (XR17D154)
//   [7:4] <<RESERVED>>
//
// UART Sleep Mode (SLEEP)
//   [0]   CH1
//   [1]   CH2
//   [2]   CH3 (XR17D154)
//   [3]   CH4 (XR17D154)
//   [7:4] <<RESERVED>>
//
// Device Revision (DREV)
//   0x01: Rev.A
//   0x02: Rev.B
//
// Device Identification (DVID)
//   0x28: XR17D158 or XR17C158
//   0x24: XR17D154 or XR17C154
//   0x22: XR17D152 or XR17C152
//   The upper nibble indicates it is a XR17xxxx serias with lower nibble
//   indicating the number of channels.
//
// REGB
//   [0]   logic 0(default) write to each UART configuration registers individually.
//         logic 1 enables simultaneous write to all 4 UARTs configuration register.
//   [3:1] <<RESERVED>>
//   [4]   Control the EECK, clock, output(pin 116) on the EEPROM interface.
//   [5]   Control the EECS, chips select, output(pin 115) to the EEPROM device.
//   [6]   EEDI(pin 114) data input. Write data to the EEPROM device.
//   [7]   EEDO(pin 113) data output. Read data from the EEPROM device.
//
// Multi-Purpose Input and Output
// MPIO interrupt mask (MPIOINT)
//   [7:0]  MPIO7～MPIO0
//   Enable MPIOx Interrupts.
//   1:enable, 0:disable(default)
//
// MPIO level control (MPIOLVL)
//   [7:0]  MPIO7～MPIO0
//   Output MPIOx pin level control.
//   1:high, 0:low(default)
//
// MPIO output control (MPIO3T)
//   [7:0]  MPIO7～MPIO0
//   Output MPIOx pin tri-state control.
//   1:tri-state, 0:active(default)
//
// MPIO input polarity select (MPIOINV)
//   [7:0]  MPIO7～MPIO0
//   Input MPIOx inversion control.
//   1:invert, 0:not invert(default)
//
// MPIO select (MPIOSEL)
//   [7:0]  MPIO7～MPIO0
//   Multipurpose input/output MPIOx pin select.
//   1:input(default), 0:output
//
typedef struct T_XR17D154_DIO_REGS {	// base+0x0080
	// SIO Interrupt
	BYTE	R_INT0;			// RO, Interrupt[3:0], Reserved[7:4]
	BYTE	R_INT1;			// RO, Interrupt Source[7:0]
	BYTE	R_INT2;			// RO, Interrupt Source[3:0], Reserved[7:4]
	BYTE	R_INT3;			// <<Reserved>>
	
	// Timer
	BYTE	R_TIMERCNTL;	// RW, Timer Control
	BYTE	R_TIMER;		// <<Reserved>>
	BYTE	R_TIMERLSB;		// RW, Timer LSB
	BYTE	R_TIMERMSB;		// RW, Timer MSB
	
	// REGA
	BYTE	R_8XMODE;		// RW
	BYTE	R_REGA;			// <<Reserved>>
	BYTE	R_RESET;		// WO, Self clear bits after executing Reset[3:0]
	BYTE	R_SLEEP;		// RW, Sleep mode[3:0]
	
	// REGB
	BYTE	R_DREV;			// RO, Device revision
	BYTE	R_DVID;			// RO, Device identification
	BYTE	R_REGB;			// RW
	
	// MPIO
	BYTE	R_MPIOINT;		// RW, MPIO interrupt mask
	BYTE	R_MPIOLVL;		// RW, MPIO level control
	BYTE	R_MPIO3T;		// RW, MPIO output control
	BYTE	R_MPIOINV;		// RW, MPIO input polarity select
	BYTE	R_MPIOSEL;		// RW, MPIO select
	
} T_XR17D154_DIO_REGS;

//------------------------------------------------------------------------------
// XR17D354/352 DIO ハードウェアレジスタ構造体
//------------------------------------------------------------------------------
// Addr   Register      Read/Write Comment
// ------ ------------- --------------------------------------------------------
// +0080: INT0[7:0]     RO, Interrupt[3:0], Reserved[7:4]
// +0081: INT1[15:8]    RO, Interrupt Source[7:0]
// +0082: INT2[23:16]   RO, Interrupt Source[3:0], Reserved[7:4]
// +0083: INT3[31:24]   <<RESERVED>>
// ------ ------------- --------------------------------------------------------
// +0084: TIMERCNTL     RW, Timer Control (※)
// +0085: TIMER         <<RESERVED>>
// +0086: TIMERLSB      RW, Timer LSB [7:0]
// +0087: TIMERMSB      RW, Timer MSB [15:8]
// ------ ------------- --------------------------------------------------------
// +0088: 8XMODE        RW, UART channel 8X Clock mode enable[3:0]
// +0089: 4XMODE        RW, UART channel 4X Clock mode enable[3:0] (※)
// +008A: RESET         WO, Self clear bits after executing Reset[3:0]
// +008B: SLEEP         RW, Sleep mode[3:0]
// ------ ------------- --------------------------------------------------------
// +008C: DREV          RO, Device revision
// +008D: DVID          RO, Device identification
// +008E: REGB          RW, configuration, EEPROM control (※)
// ------ ------------- --------------------------------------------------------
// +008F: MPIOINT[7:0]  RW, MPIO interrupt mask
// +0090: MPIOLVL[7:0]  RW, MPIO level control
// +0091: MPIO3T[7:0]   RW, MPIO output control
// +0092: MPIOINV[7:0]  RW, MPIO input polarity select
// +0093: MPIOSEL[7:0]  RW, MPIO select
// +0094: MPIOOD[7:0]   RW, MPIO open-drain output control (※)
// ------ ------------- --------------------------------------------------------
// +0095: MPIOINT[15:8] RW, MPIO interrupt mask (※)
// +0096: MPIOLVL[15:8] RW, MPIO level control (※)
// +0097: MPIO3T[15:8]  RW, MPIO output control (※)
// +0098: MPIOINV[15:8] RW, MPIO input polarity select (※)
// +0099: MPIOSEL[15:8] RW, MPIO select (※)
// +009A: MPIOOD[15:8]  RW, MPIO open-drain output control (※)
// ------ ------------- --------------------------------------------------------
// +009B: <<RESERVED>>
// ------ ------------- --------------------------------------------------------
//
// SIO Interrupt (INT0)
//   [0]   CH1
//   [1]   CH2
//   [2]   CH3
//   [3]   CH4
//   [7:4] <<RESERVED>>
//
// SIO Interrupt source (INT2[3:0]+INT1[7:0])
//   [ 2: 0](INT1[2:0])         CH1 Interrupt Source Encoding
//   [ 5: 3](INT1[5:3])         CH2 Interrupt Source Encoding
//   [ 8: 6](INT2[0]+INT1[7:6]) CH3 Interrupt Source Encoding
//   [11: 9](INT2[3:1])         CH4 Interrupt Source Encoding
//   [15:12](INT2[7:4])         <<RESERVED>>
//
// Interrupt Encoding(0～7)
//   0: None or wake-up indicator (wake-up indicator is reported in channel 0 only).
//   1: RXRDY and RX line Status (logic OR of LSR[4:1])
//   2: RXRDY Time-out
//   3: TXRDY, THR or TSR(auto RS485 mode) empty
//   4: MSR, RTS/CTS or DTR/DSR delta or Xoff/Xon det. or special char. detected.
//   5: <<Reserved>>
//   6: MPIO pin(s). Reported in channel 0 only.
//   7: TIMER time-out. Reported in channel 0 only.
//
//   RXRDY is cleared by reading data in th RX FIFO until it falls below the trigger level.
//   RXRDY Time-out is cleared by reading data until the RX FIFO is empty.
//   RX Line Status interrupt clears after reading the LSR register.
//   TXRDY interrupt clears after reading ISR register that is in the UART channel register set.
//   Modem Status Register interrupt clears after reading MSR register that is in the UART channel register set.
//   RTS/CTS or DTR/DSR delta interrupt clears after reading MSR register that is in the UART channel register set.
//   Xoff/Xon interrupt clears after reading the ISR register that is in the UART channel register set.
//   Special character detect interrupt is cleared by read to ISR or after the next character is received.
//   TIMER Time-out interrupt clears after reading the TIMERCNTL register that is in the Device Configuration register set.
//   MPIO interrupt clears after reading MPIOLVL register that is the Device Configuration register set.
//
// Timer Control (TIMERCNTL)
//   [3:0] commands
//         0001: Enable Timer Interrupt
//         0010: Disable Timer Interrupt
//         0011: Select One-shot mode
//         0100: Select Re-triggerable mode
//         0101: Select Internal 125MHz clock as clock input for the Timer
//         0110: Select External Clock input through the TMRCK pin for the Timer
//         0111: Route Timer output to MPIO[0] pin
//         1000: De-route Timer output from MPIO[0]
//         1001: Start Timer
//         1010: Stop Timer
//         1011: Reset Timer
//         1100～1111: <<RESERVED>
//   [7:4] <<RESERVED>>
//
// 8X Clock Enable (8XMODE)
//   [0]   CH1
//   [1]   CH2
//   [2]   CH3
//   [3]   CH4
//   [7:4] <<RESERVED>>
//
// 4X Clock Enable (4XMODE)
//   [0]   CH1
//   [1]   CH2
//   [2]   CH3
//   [3]   CH4
//   [7:4] <<RESERVED>>
//
// UART Reset (RESET)
//   [0]   CH1
//   [1]   CH2
//   [2]   CH3
//   [3]   CH4
//   [7:4] <<RESERVED>>
//
// UART Sleep Mode (SLEEP)
//   [0]   CH1
//   [1]   CH2
//   [2]   CH3
//   [3]   CH4
//   [7:4] <<RESERVED>>
//
// Device Revision (DREV)
//   0x01: Rev.A
//   0x02: Rev.B
//
// Device Identification (DVID)
//   0x84: XR17D354
//   0x82: XR17D352
//   The upper nibble indicates it is a XR17xxxx serias with lower nibble
//   indicating the number of channels.
//
// REGB
//   [0]   logic 0(default) write to each UART configuration registers individually.
//         logic 1 enables simultaneous write to all 4 UARTs configuration register.
//   [1]   logic 0(default) wake-up interrupt is generated when UART exits sleep mode.
//         logic 1 No wake-up interrupt is generated when UART exits sleep mode.
//   [2]   logic 0(default) Global interrupt enable. Interrupts to PCI host are enabled.
//         logic 1 Global interrupt disable. Interrupts to PCI host are disabled.
//   [3]   logic 0(default) EEPROM load is valid.
//         logic 1 EEPROM load error caused by on of the following conditions: EEPROM not attached, final bit not found, parity error detected.
//   [4]   Control the EECK, clock, output(pin 116) on the EEPROM interface.
//   [5]   Control the EECS, chips select, output(pin 115) to the EEPROM device.
//   [6]   EEDI(pin 114) data input. Write data to the EEPROM device.
//   [7]   EEDO(pin 113) data output. Read data from the EEPROM device.
//
// Multi-Purpose Input and Output
// MPIO interrupt mask (MPIOINT)
//   [7:0]  MPIO7～MPIO0
//   [15:8] MPIO15～MPIO8
//   Enable MPIOx Interrupts.
//   1:enable, 0:disable(default)
//
// MPIO level control (MPIOLVL)
//   [7:0]  MPIO7～MPIO0
//   [15:8] MPIO15～MPIO8
//   Output MPIOx pin level control.
//   1:high, 0:low(default)
//
// MPIO output control (MPIO3T)
//   [7:0]  MPIO7～MPIO0
//   [15:8] MPIO15～MPIO8
//   Output MPIOx pin tri-state control.
//   1:tri-state, 0:active(default)
//
// MPIO input polarity select (MPIOINV)
//   [7:0]  MPIO7～MPIO0
//   [15:8] MPIO15～MPIO8
//   Input MPIOx inversion control.
//   1:invert, 0:not invert(default)
//
// MPIO select (MPIOSEL)
//   [7:0]  MPIO7～MPIO0
//   [15:8] MPIO15～MPIO8
//   Multipurpose input/output MPIOx pin select.
//   1:input(default), 0:output
//
// MPIO open-drain output control (MPIOOD)
//   [7:0]  MPIO7～MPIO0
//   [15:8] MPIO15～MPIO8
//   1:open-drain output, 0: normal
//
typedef struct T_XR17D354_DIO_REGS {	// base+0x0080
	// SIO Interrupt
	BYTE	R_INT0;			// RO, Interrupt[3:0], Reserved[7:4]
	BYTE	R_INT1;			// RO, Interrupt Source[7:0]
	BYTE	R_INT2;			// RO, Interrupt Source[3:0], Reserved[7:4]
	BYTE	R_INT3;			// <<Reserved>>
	
	// Timer
	BYTE	R_TIMERCNTL;	// RW, Timer Control
	BYTE	R_TIMER;		// <<Reserved>>
	BYTE	R_TIMERLSB;		// RW, Timer LSB
	BYTE	R_TIMERMSB;		// RW, Timer MSB
	
	// REGA
	BYTE	R_8XMODE;		// RW
	BYTE	R_REGA;			// <<Reserved>>
	BYTE	R_RESET;		// WO, Self clear bits after executing Reset[3:0]
	BYTE	R_SLEEP;		// RW, Sleep mode[3:0]
	
	// REGB
	BYTE	R_DREV;			// RO, Device revision
	BYTE	R_DVID;			// RO, Device identification
	BYTE	R_REGB;			// RW
	
	// MPIO
	BYTE	R_MPIOINT;		// RW, MPIO[7:0] interrupt mask
	BYTE	R_MPIOLVL;		// RW, MPIO[7:0] level control
	BYTE	R_MPIO3T;		// RW, MPIO[7:0] output control
	BYTE	R_MPIOINV;		// RW, MPIO[7:0] input polarity select
	BYTE	R_MPIOSEL;		// RW, MPIO[7:0] select
	BYTE	R_MPIOOD;		// RW, MOIO[7:0] open drain control
	
	BYTE	R_MPIOINT_H;	// RW, MPIO[15:8] interrupt mask
	BYTE	R_MPIOLVL_H;	// RW, MPIO[15:8] level control
	BYTE	R_MPIO3T_H;		// RW, MPIO[15:8] output control
	BYTE	R_MPIOINV_H;	// RW, MPIO[15:8] input polarity select
	BYTE	R_MPIOSEL_H;	// RW, MPIO[15:8] select
	BYTE	R_MPIOOD_H;		// RW, MOIO[15:8] open drain control
} T_XR17D354_DIO_REGS;


//==============================================================================
// ドライバ変数領域定義
//==============================================================================

//------------------------------------------------------------------------------
// 割り込みシグナル設定情報テーブル
//------------------------------------------------------------------------------
typedef struct T_DEVICE_SIGINFO {
	struct file				*pFile;			// シグナルを設定した時の struct file のポインタ,
	int						nSignal;		// シグナル番号
	struct task_struct		*pTask;			// シグナル送信先タスク
} T_DEVICE_SIGINFO;

//------------------------------------------------------------------------------
// DIO関係レジスタの初期設定値テーブル定義
//------------------------------------------------------------------------------
typedef struct T_DIO_REG_INIT_VALUE {
	BYTE	MPIOINT;	// [7:0] MPIO7～0, 1:enable, 0:disable(default)
	BYTE	MPIOLVL;	// [7:0] MPIO7～0, 1:highm 0:low(default)
	BYTE	MPIO3T;		// [7:0] MPIO7～0, 1:tri-state, 0:active(default)
	BYTE	MPIOINV;	// [7:0] MPIO7～0, 1:invert, 0:not invert(default)
	BYTE	MPIOSEL;	// [7:0] MPIO7～0, 1:input(default), 0:output
	BYTE	MPIOOD;		// [7:0] MPIO7～0, 1:open-drain, 0:normal(default) ※未使用
} T_DIO_REG_INIT_VALUE;

//------------------------------------------------------------------------------
// DIOドライバの初期設定情報テーブル定義
//------------------------------------------------------------------------------
typedef struct T_IDS_DRIVER_DATA {
	T_DIO_REG_INIT_VALUE	sInitREG;		// レジスタ初期設定値[7:0]用
	T_DIO_REG_INIT_VALUE	sInitREGH;		// レジスタ初期設定値[15:8]用 (XR17D35xのみ使用)
	WORD					wUserMask;		// ユーザ変更可能なMPIO[15:0]を示すマスク(1:user, 0:reserved)
	int		(*pfnInitHW)(struct T_IDS_DRIVER_DATA *pData, void __iomem *pBase);		// ハードウェア初期化処理
	int		(*pfnDeinitHW)(struct T_IDS_DRIVER_DATA *pData, void __iomem *pBase);	// ハードウェア後始末処理
	int		(*pfnReadDipsw)(struct T_IDS_DRIVER_DATA *pData, void __iomem *pBase);	// DIPSW読み出し処理
} T_IDS_DRIVER_DATA;

//------------------------------------------------------------------------------
// デバイス固有変数テーブル
//------------------------------------------------------------------------------
typedef struct T_DEVICE_STORAGE {
	struct semaphore		semDevice;		// exclution semaphore
	BOOL					bEnable;		// デバイス有効フラグ(probe時にset)
	
	T_IDS_DRIVER_DATA		*pIDS;			// ドライバ初期化データ
	
	int						nSubSysID;		// サブシステムID(0x6005～0x600C)
	int						nID;			// ID番号(0～3、DIPSWの設定と同じ値)
	struct pci_dev			*pDev;			// pci_dev構造体へのポインタ
	struct device			*pDevice;		// クラスデバイス
	
	int						nIRQ;			// PCI割込みライン
	void __iomem			*pBase;			// レジスタのベースアドレス(BAR0のベースをioremap()したもの)
	void __iomem			*pREGs;			// DIOレジスタアクセス用ポインタ(T_XR17D154_DIO_REGSの先頭アドレス)
	
	BYTE					uUserMask;		// ユーザ利用可能なMPIO[7:0]のマスク(1:user, 0:reserved)
	BYTE					uUserMaskH;		// ユーザ利用可能なMPIO[15:8]のマスク(1:user, 0:reserved)
	
	T_DEVICE_SIGINFO		sSigInfo;		// シグナル情報
#if 1
	struct semaphore		semINT;			// 割り込み通知用セマフォ(割り込みハンドラ⇒割り込み処理タスク)
	struct task_struct		*pIntTask;		// 割り込み処理タスクの情報
	BOOL					bKillIntTask;	// 割り込み処理タスクを停止させる時にTRUE
	
	struct semaphore		semIoctlINT;	// ioctl(IOCTL_DIO_START_INTERRUPT)で待つセマフォ(割り込み処理タスク⇒ioctl)
	int						nIntStat;		// 割り込みステータス(1:割り込み有り, 0:割り込み無し)
	BOOL					bWaitIoctl;		// ioctl(IOCTL_DIO_START_INTERRUPT)で待ちならTRUE
#else
	struct completion		compINT;		// 割り込み通知用completion(割り込みハンドラ⇒割り込み処理タスク)
	struct task_struct		*pIntTask;		// 割り込み処理タスクの情報
	BOOL					bKillIntTask;	// 割り込み処理タスクを停止させる時にTRUE
	
	struct completion		compIntWait;	// ioctl(IOCTL_DIO_START_INTERRUPT)で待つセマフォ(割り込み処理タスク⇒ioctl)
	int						nIntStat;		// 割り込みステータス(1:割り込み有り, 0:割り込み無し)
	BOOL					bWaitIoctl;		// ioctl(IOCTL_DIO_START_INTERRUPT)で待ちならTRUE
#endif
	struct timer_list		sIntTimer;		// ioctl()で割込みを待つタイムアウト時間を管理するタイマ(10秒)
	
} T_DEVICE_STORAGE;


//------------------------------------------------------------------------------
// ドライバ固有変数テーブル
//------------------------------------------------------------------------------
typedef struct T_DRIVER_STORAGE {
	// デバイス共通
	int						nUsed;			// ドライバ使用カウンタ
	int						nMajor;			// ドライバメジャー番号
	struct class			*psClass;		// デバイスクラス
	
	// デバイス毎(DIPSWによる#0～#3の4台分を管理)
	T_DEVICE_STORAGE		sDevice[REX_DIO_MAX_BOARDS];
	
} T_DRIVER_STORAGE;

//------------------------------------------------------------------------------
// 唯一のドライバ固有変数テーブルへのポインタ
//------------------------------------------------------------------------------
static T_DRIVER_STORAGE	*s_psDriver = NULL;

/*
//------------------------------------------------------------------------------
// ドライバ関数プロトタイプ宣言
//------------------------------------------------------------------------------

// ドライバエントリ初期化処理
// static int rexpci60d_probe(struct pci_dev *link, const struct pci_device_id *id);
//   link     (IN)   pci_dev構造体のポインタ
//   id       (IN)   pci_device_id構造体のポインタ
//   <RETURN>       0:OK
static int rexpci60d_probe(struct pci_dev *link, const struct pci_device_id *id);

// ドライバエントリ終了処理
// static void rexpci60d_remove(struct pci_dev *link);
//   link     (IN)   pci_dev構造体のポインタ
static void rexpci60d_remove(struct pci_dev *link);

// インタラプト転送処理
//static void rexpci60d_isr(int irq, void *dev_id, struct pt_regs *regs);
//   irq      (IN)  割込み番号
//   dev_id   (IN)  rs_dio構造体のポインタ(user設定用パラメタ)
//   regs     (IN)  pt_regs構造体のポインタ
static void rexpci60d_isr(int irq, void *dev_id, struct pt_regs *regs);

// デバイスの初期化処理
// static int dio_open(struct inode *inode, struct file *file);
//   inode    (IN)  inode構造体のポインタ
//   file     (IN)  file構造体のポインタ
//   <RETURN>       0:OK, 0未満:エラーコード
static int dio_open(struct inode *inode, struct file *file);

// デバイスの終了処理
// static int dio_release(struct inode *inode, struct file *file);
//   inode    (IN)  inode構造体のポインタ
//   file     (IN)  file構造体のポインタ
//   <RETURN>       0:OK, 0未満:エラーコード
static int dio_release(struct inode *inode, struct file *file);

// IOCTL処理
// static int dio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
//   inode    (IN)  inode構造体のポインタ
//   file     (IN)  file構造体のポインタ
//   cmd      (IN)  コントロールコード
//   arg      (IN)  付加情報
//   <RETURN>       0:OK, 0未満:エラーコード
static int dio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);

// ローダブルモジュール初期化処理
// static int __init rexpci60d_init_module(void);
//   <RETURN>       0:OK, !0:NG
static int __init rexpci60d_init_module(void);

// ローダブルモジュール終了処理
// static void __exit rexpci60d_exit_module(void);
static void __exit rexpci60d_exit_module(void);
*/

//==============================================================================
// キャラクタデバイスのファイル操作エントリ
//==============================================================================

typedef struct T_FILE_PRIVATE {
	int			nDevNo;			// デバイス番号(0～3)
} T_FILE_PRIVATE;

//------------------------------------------------------------------------------
// デバイスの初期化処理
// static int dio_open(struct inode *inode, struct file *file);
//   inode    (IN)  inode構造体のポインタ
//   file     (IN)  file構造体のポインタ
//   <RETURN>       0:OK, 0未満:エラーコード
//------------------------------------------------------------------------------
static int dio_open(struct inode *inode, struct file *file)
{
	T_FILE_PRIVATE	*pPrivate;
	int				nReturn;
	
	pPrivate = kmalloc(sizeof(T_FILE_PRIVATE), GFP_KERNEL);
	if (pPrivate == NULL) {
		nReturn = -ENOMEM;
	}
	else {
		pPrivate->nDevNo = iminor(inode);		// マイナー番号がID番号(0～3)
		file->private_data = pPrivate;
		nReturn = 0;
	}
	
	return nReturn;
}

//------------------------------------------------------------------------------
// デバイスの終了処理
// static int dio_release(struct inode *inode, struct file *file);
//   inode    (IN)  inode構造体のポインタ
//   file     (IN)  file構造体のポインタ
//   <RETURN>       0:OK, 0未満:エラーコード
//------------------------------------------------------------------------------
static int dio_release(struct inode *inode, struct file *file)
{
	int					nDevNo;
	T_DEVICE_STORAGE	*pStorage;
	
	nDevNo = iminor(inode);		// マイナー番号がID番号(0～3)
	pStorage = &s_psDriver->sDevice[nDevNo];
	
	down(&pStorage->semDevice);
	if (pStorage->sSigInfo.pFile == file && pStorage->sSigInfo.pTask == current) {
		// 割り込みシグナルを使用中のハンドルが閉じられる場合は割り込みシグナル設定を解除
		T_XR17D154_DIO_REGS	*pREGs = pStorage->pREGs;
		
		writeb(0x00, &pREGs->R_MPIOINT);	// MPIO割り込み無効
		
		pStorage->sSigInfo.pFile = NULL;
		pStorage->sSigInfo.pTask = NULL;
		pStorage->sSigInfo.nSignal = 0;
	}
	up(&pStorage->semDevice);
	
	if (file->private_data != NULL) {
		kfree(file->private_data);
	}
	
	return 0;
}

//------------------------------------------------------------------------------
// タイムアウトハンドラ
//------------------------------------------------------------------------------
static void dio_timeout_handler(unsigned long ulData)
{
	DEBUG_PRINTK(KERN_NOTICE DRIVER_NAME ": dio_timeout_handler(%08lX)\n", ulData);
	
	s_psDriver->sDevice[ulData].bWaitIoctl = FALSE;	// 待ち状態を解除
	up(&s_psDriver->sDevice[ulData].semIoctlINT);	// ioctl()の割り込みイベント待ちを解除
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
//------------------------------------------------------------------------------
// IOCTL処理 (kernel 2.6.35 以前)
// static int dio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
//   inode    (IN)  inode構造体のポインタ
//   file     (IN)  file構造体のポインタ
//   cmd      (IN)  コントロールコード
//   arg      (IN)  付加情報
//   <RETURN>       0:OK, 0未満:エラーコード
//
// ※引数にinodeを渡す古い形式の ioctl は kernel 2.6.35 を最後に廃止された。
//------------------------------------------------------------------------------
static int dio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
#else
//------------------------------------------------------------------------------
// IOCTL処理 (kernel 2.6.36 以降)
// static int dio_ioctl(struct file *file, unsigned int cmd, unsigned long arg);
//   file     (IN)  file構造体のポインタ
//   cmd      (IN)  コントロールコード
//   arg      (IN)  付加情報
//   <RETURN>       0:OK, 0未満:エラーコード
//
// ※kernel 2.6.36以降は 2.6.11 から追加された unlocked_ioctl や compat_ioctl を
//   古い ioctl の代わりに使用する。
//   なお、通常は ioctl の置き換えは unlocked_ioctl を用いる。
//------------------------------------------------------------------------------
static long dio_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#endif
{
	int		nError = 0;
	int		nReturn = -EINVAL;
	int		nDevNo;
	
	DEBUG_PRINTK(KERN_NOTICE DRIVER_NAME ": dio_ioctl(%08lX,%08X,%08lX)\n", (unsigned long)file, cmd, arg);
	
	if (_IOC_TYPE(cmd) != REX_DIO_IOC_MAGIC) {
		// マジックコードが不一致 ⇒ 無効なioctl番号
		return -ENOTTY;
	}
	
	if (_IOC_NR(cmd) > REX_DIO_IOC_MAXNR) {
		// 機能番号が範囲外 ⇒ 無効なioctl番号
		return -ENOTTY;
	}
	
	if (_IOC_DIR(cmd) & _IOC_READ) {
		// read方向のioctl (ドライバから見るとユーザ空間へ書き込むのでWRITEチェック)
		nError = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
	}
	if (_IOC_DIR(cmd) & _IOC_WRITE) {
		// write方向のioctl (ドライバから見るとユーザ空間から読み出すのでREADチェック)
		nError = !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
	}
	if (nError != 0) {
		return -EFAULT;
	}
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
	nDevNo = iminor(inode);		// マイナー番号がID番号(0～3)
#else
	if (file->private_data == NULL) {
		return -EFAULT;
	}
	else {
		T_FILE_PRIVATE	*pPrivate;
		
		pPrivate = (T_FILE_PRIVATE *)file->private_data;
		
		nDevNo = pPrivate->nDevNo;
	}
#endif
	if (0 <= nDevNo && nDevNo < REX_DIO_MAX_BOARDS) {
		switch (cmd) {
		case IOCTL_DIO_READ_PORT:
			// PORTからの読み出し
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				IOPORT_UCHAR	sParam;
				
				nReturn = copy_from_user(&sParam, (void __user *)arg, sizeof(sParam));
				if (nReturn == 0) {
					if (sParam.Offset < 0x80 || 0xFF < sParam.Offset) {
						nReturn = -EINVAL;
					}
					else {
						BYTE	*pBase = (BYTE *)s_psDriver->sDevice[nDevNo].pBase;
						
						sParam.Value = readb(&pBase[sParam.Offset]);
						nReturn = copy_to_user((void __user *)arg, &sParam, sizeof(sParam));
					}
				}
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			break;
		case IOCTL_DIO_WRITE_PORT:
			// PORTへの書込み
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				IOPORT_UCHAR	sParam;
				
				nReturn = copy_from_user(&sParam, (void __user *)arg, sizeof(sParam));
				if (nReturn == 0) {
					if (sParam.Offset < 0x80 || 0xFF < sParam.Offset) {
						nReturn = -EINVAL;
					}
					else {
						BYTE	*pBase = (BYTE *)s_psDriver->sDevice[nDevNo].pBase;
						
#if 1
						switch (sParam.Offset) {
//						case 0x92:	// MPIOINV
						case 0x93:	// MPIOSEL
							{
								BYTE	uMask;
								BYTE	uData;
								
								uMask = s_psDriver->sDevice[nDevNo].uUserMask;
								uData = (sParam.Value & uMask) | (readb(&pBase[sParam.Offset]) & ~uMask);
								writeb(uData, &pBase[sParam.Offset]);
							}
							break;
						default:
							writeb(sParam.Value, &pBase[sParam.Offset]);
						}
#else
						writeb(sParam.Value, &pBase[sParam.Offset]);
#endif
					}
				}
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			break;
		case IOCTL_DIO_START_INTERRUPT:
			// 割り込み待ち開始
#if 1
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				// ioctlの割り込みイベント待ち用セマフォをLOCK状態にする
				while (down_trylock(&s_psDriver->sDevice[nDevNo].semIoctlINT) == 0) ;
				
				s_psDriver->sDevice[nDevNo].nIntStat = 0;
				s_psDriver->sDevice[nDevNo].bWaitIoctl = TRUE;
				
				s_psDriver->sDevice[nDevNo].sIntTimer.expires = jiffies + (10*HZ);	// timeout = 10s
				s_psDriver->sDevice[nDevNo].sIntTimer.data = nDevNo;
				s_psDriver->sDevice[nDevNo].sIntTimer.function = dio_timeout_handler;
				add_timer(&s_psDriver->sDevice[nDevNo].sIntTimer);
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			
			// ioctlの割り込みイベント待ち用セマフォが解除されるのを待つ
			if (down_interruptible(&s_psDriver->sDevice[nDevNo].semIoctlINT) != 0) {
				// down()が割り込まれたら中止
				nReturn = -ECANCELED;
				down(&s_psDriver->sDevice[nDevNo].semDevice);
				del_timer_sync(&s_psDriver->sDevice[nDevNo].sIntTimer);
				s_psDriver->sDevice[nDevNo].bWaitIoctl = FALSE;
				up(&s_psDriver->sDevice[nDevNo].semDevice);
			}
			else {
				down(&s_psDriver->sDevice[nDevNo].semDevice);
				del_timer_sync(&s_psDriver->sDevice[nDevNo].sIntTimer);
				s_psDriver->sDevice[nDevNo].bWaitIoctl = FALSE;
				if (s_psDriver->sDevice[nDevNo].nIntStat > 0) {
					// DIO割り込みがあった
					nReturn = 0;
				}
				else {
					// DIO割り込みが無かったら中止されたという事
					nReturn = -ECANCELED;
				}
				up(&s_psDriver->sDevice[nDevNo].semDevice);
			}
#else
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				// IntWaitのセマフォをLOCK状態にする
				INIT_COMPLETION(&s_psDriver->sDevice[nDevNo].compIntWait);
				
				s_psDriver->sDevice[nDevNo].nIntStat = 0;
				s_psDriver->sDevice[nDevNo].bWaitIoctl = TRUE;
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			
			// 割り込み通知セマフォが解除されるのを待つ
			wait_for_completion(&s_psDriver->sDevice[nDevNo].compIntWait);
			
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				s_psDriver->sDevice[nDevNo].bWaitIoctl = FALSE;
				if (s_psDriver->sDevice[nDevNo].nIntStat > 0) {
					// DIO割り込みがあった
					nReturn = 0;
				}
				else {
					// DIO割り込みが無かったら中止されたという事
					nReturn = -ECANCELED;
				}
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
#endif
			break;
		case IOCTL_DIO_STOP_INTERRUPT:
			// 割り込み待ち終了
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				s_psDriver->sDevice[nDevNo].bWaitIoctl = FALSE;	// 待ち状態を解除
#if 1
				up(&s_psDriver->sDevice[nDevNo].semIoctlINT);	// 中止のためのイベント
#else
				complete(&s_psDriver->sDevice[nDevNo].compIntWait);
#endif
				nReturn = 0;
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			break;
		case REX_DIO_IOCG_READ:
			// 入力ポートからの読み出し
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				T_IOCTL_DIO_READ	sParam;
				T_XR17D154_DIO_REGS	*pREGs = s_psDriver->sDevice[nDevNo].pREGs;
				
				sParam.ucData = readb(&pREGs->R_MPIOLVL) & s_psDriver->sDevice[nDevNo].uUserMask;
				nReturn = copy_to_user((void __user *)arg, &sParam, sizeof(sParam));
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			break;
		case REX_DIO_IOCS_WRITE:
			// 出力ポートへの書込み
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				T_IOCTL_DIO_WRITE	sParam;
				
				nReturn = copy_from_user(&sParam, (void __user *)arg, sizeof(sParam));
				if (nReturn == 0) {
					BYTE				uData;
					BYTE				uMask;
					T_XR17D154_DIO_REGS	*pREGs = s_psDriver->sDevice[nDevNo].pREGs;
					
					uMask = s_psDriver->sDevice[nDevNo].uUserMask;
					
					uData = (sParam.ucData & uMask) | (readb(&pREGs->R_MPIOLVL) & ~uMask);
					writeb(uData, &pREGs->R_MPIOLVL);
				}
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			break;
		case REX_DIO_IOCS_INTSIG_ENABLE:
			// 割り込みシグナル設定
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				T_IOCTL_DIO_INT_ENABLE	sParam;
				
				nReturn = copy_from_user(&sParam, (void __user *)arg, sizeof(sParam));
				if (nReturn == 0) {
					T_XR17D154_DIO_REGS	*pREGs = s_psDriver->sDevice[nDevNo].pREGs;
					
					if ((sParam.ucIntBit & (~readb(&pREGs->R_MPIOSEL))) != 0) {
						// 入力でないポートに割り込みを設定しようとした
						nReturn = -EINVAL;		// -1
					}
					else if (s_psDriver->sDevice[nDevNo].sSigInfo.nSignal != 0) {
						// 既に登録されていたらエラー
						nReturn = -EALREADY;	// -4
					}
					else {
						BYTE	uIntData;
						BYTE	uMask;
						
						s_psDriver->sDevice[nDevNo].sSigInfo.pFile = file;
						s_psDriver->sDevice[nDevNo].sSigInfo.nSignal = sParam.nSignalNo;
						s_psDriver->sDevice[nDevNo].sSigInfo.pTask = current;
						
						uMask = s_psDriver->sDevice[nDevNo].uUserMask;
						uIntData = (sParam.ucIntBit & uMask) | (readb(&pREGs->R_MPIOINT) & ~uMask);
						writeb(uIntData, &pREGs->R_MPIOINT);
					}
				}
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			break;
		case REX_DIO_IOCS_INTSIG_DISABLE:
			// 割り込みシグナル解除
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				T_IOCTL_DIO_INT_DISABLE	sParam;
				
				nReturn = copy_from_user(&sParam, (void __user *)arg, sizeof(sParam));
				if (nReturn == 0) {
					T_XR17D154_DIO_REGS	*pREGs = s_psDriver->sDevice[nDevNo].pREGs;
					
					if (s_psDriver->sDevice[nDevNo].sSigInfo.pFile == file
					 && s_psDriver->sDevice[nDevNo].sSigInfo.nSignal == sParam.nSignalNo
					 && s_psDriver->sDevice[nDevNo].sSigInfo.pTask == current) {
						// ファイルハンドルとシグナル番号とタスクが一致する場合のみ処理する
						writeb(0x00, &pREGs->R_MPIOINT);	// MPIO割り込み無効
						
						s_psDriver->sDevice[nDevNo].sSigInfo.pFile = NULL;
						s_psDriver->sDevice[nDevNo].sSigInfo.pTask = NULL;
						s_psDriver->sDevice[nDevNo].sSigInfo.nSignal = 0;
					}
					else {
						nReturn = -EINVAL;
					}
				}
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			break;
		case REX_DIO_IOCS_DIRECTION:
			// 入出力方向設定
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				T_IOCTL_DIO_DIRECTION	sParam;
				
				nReturn = copy_from_user(&sParam, (void __user *)arg, sizeof(sParam));
				if (nReturn == 0) {
					BYTE				uData;
					BYTE				uMask;
					T_XR17D154_DIO_REGS	*pREGs = s_psDriver->sDevice[nDevNo].pREGs;
					
					uMask = s_psDriver->sDevice[nDevNo].uUserMask;
					uData = (sParam.ucDirection & uMask) | (readb(&pREGs->R_MPIOSEL) & ~uMask);
					writeb(uData, &pREGs->R_MPIOSEL);
				}
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			break;
		case REX_DIO_IOCS_INVERT:
			// INVERT設定
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				T_IOCTL_DIO_INVERT	sParam;
				
				nReturn = copy_from_user(&sParam, (void __user *)arg, sizeof(sParam));
				if (nReturn == 0) {
					BYTE				uData;
					BYTE				uMask;
					T_XR17D154_DIO_REGS	*pREGs = s_psDriver->sDevice[nDevNo].pREGs;
					
					uMask = s_psDriver->sDevice[nDevNo].uUserMask;
					uData = (sParam.ucInvert & uMask) | (readb(&pREGs->R_MPIOINV) & ~uMask);
					writeb(uData, &pREGs->R_MPIOINV);
				}
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			break;
		case REX_DIO_IOCS_TRISTATE:
			// TRI-STATE設定
			down(&s_psDriver->sDevice[nDevNo].semDevice);
			{
				T_IOCTL_DIO_TRISTATE	sParam;
				
				nReturn = copy_from_user(&sParam, (void __user *)arg, sizeof(sParam));
				if (nReturn == 0) {
					BYTE				uData;
					BYTE				uMask;
					T_XR17D154_DIO_REGS	*pREGs = s_psDriver->sDevice[nDevNo].pREGs;
					
					uMask = s_psDriver->sDevice[nDevNo].uUserMask;
					uData = (sParam.ucTriState & uMask) | (readb(&pREGs->R_MPIO3T) & ~uMask);
					writeb(uData, &pREGs->R_MPIO3T);
				}
			}
			up(&s_psDriver->sDevice[nDevNo].semDevice);
			break;
		default:
			nReturn = -ENOTTY;
		}
	}
	
	return nReturn;
}


//------------------------------------------------------------------------------
// file_operations 構造体
//------------------------------------------------------------------------------
static struct file_operations dio_fops = {
	.owner			= THIS_MODULE,
	.open			= dio_open,
	.release		= dio_release,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
	.ioctl			= dio_ioctl,
#else
	.unlocked_ioctl	= dio_ioctl,
#endif
};


#ifdef USE_DEBUG_PROC_REXDIO
//==============================================================================
// DEBUG用 /proc/rexdio ドライバ
//==============================================================================
//static char szProcString[80*100];

//------------------------------------------------------------------------------
// /proc/rexdio read 処理(DEBUG用)
//------------------------------------------------------------------------------
int rexdio_read_proc(char *pBuffer, char **ppStart, off_t nOffset, int nCount, int *pbEOF, void *pData)
{
	int		nLen = 0;
	int		i;
	
	for (i = 0; i < REX_DIO_MAX_BOARDS; i++) {
		if (s_psDriver->sDevice[i].bEnable) {
			T_XR17D154_DIO_REGS		*pREGs;
			
			pREGs = s_psDriver->sDevice[i].pREGs;
			
			// to dmesg
			printk(KERN_NOTICE DRIVER_NAME ": BOARD ID=%d\n", s_psDriver->sDevice[i].nID);
			printk(KERN_NOTICE DRIVER_NAME ":  INT0   =%02X\n", readb(&pREGs->R_INT0));
			printk(KERN_NOTICE DRIVER_NAME ":  INT1   =%02X\n", readb(&pREGs->R_INT1));
			printk(KERN_NOTICE DRIVER_NAME ":  INT2   =%02X\n", readb(&pREGs->R_INT2));
			printk(KERN_NOTICE DRIVER_NAME ":  INT3   =%02X\n", readb(&pREGs->R_INT3));
			printk(KERN_NOTICE DRIVER_NAME ":  MPIOINT=%02X\n", readb(&pREGs->R_MPIOINT));
			printk(KERN_NOTICE DRIVER_NAME ":  MPIOLVL=%02X\n", readb(&pREGs->R_MPIOLVL));
			printk(KERN_NOTICE DRIVER_NAME ":  MPIO3 T=%02X\n", readb(&pREGs->R_MPIO3T));
			printk(KERN_NOTICE DRIVER_NAME ":  MPIOINV=%02X\n", readb(&pREGs->R_MPIOINV));
			printk(KERN_NOTICE DRIVER_NAME ":  MPIOSEL=%02X\n", readb(&pREGs->R_MPIOSEL));
		}
	}
	
	*pbEOF = 1;
	
	return nLen;
}

//------------------------------------------------------------------------------
// /proc/rexdio エントリ生成
//------------------------------------------------------------------------------
static void rexdio_create_proc(void)
{
	struct proc_dir_entry	*pEntry;
	
	pEntry = create_proc_entry(PROC_NAME, 0, 0);
	pEntry->read_proc = rexdio_read_proc;	// proc read 処理
//	pEntry->write_proc = rexdio_write_proc;	// proc write 処理
}

//------------------------------------------------------------------------------
// /proc/rexdio エントリ削除
//------------------------------------------------------------------------------
static void rexdio_remove_proc(void)
{
	remove_proc_entry(PROC_NAME, NULL);
}

#endif


//==============================================================================
// 割り込み関連
//==============================================================================

//------------------------------------------------------------------------------
// インタラプト転送処理
//static void rexpci60d_isr(int irq, void *dev_id, struct pt_regs *regs);
//   irq      (IN)  割込み番号
//   dev_id   (IN)  rs_dio構造体のポインタ(user設定用パラメタ)
//   regs     (IN)  pt_regs構造体のポインタ
//------------------------------------------------------------------------------
//static void rexpci60d_isr(int irq, void *dev_id, struct pt_regs *regs)
//static irqreturn_t rexpci60d_isr(int irq, void *dev_id, struct pt_regs *regs)
static irqreturn_t rexpci60d_isr(int irq, void *dev_id)
{
	T_DEVICE_STORAGE	*pStorage = (T_DEVICE_STORAGE *)dev_id;
	T_XR17D154_DIO_REGS	*pREGs;
	irqreturn_t			nIRQReturn = IRQ_NONE;
	DWORD				dwINT;
	
	pREGs = pStorage->pREGs;
	
	dwINT = readl(&pREGs->R_INT0);
	
#ifdef USE_DEBUG_IRQ_HANDLER
	{
		volatile BYTE	ucData;
		
		// 割り込み解除
		ucData = readb(&pREGs->R_MPIOLVL);	// Clear IRQ
		writeb(0x00, &pREGs->R_MPIOINT);	// MPIO割り込みを無効にする
		
# if 1
		// 割り込み通知
		up(&pStorage->semINT);
# else
		// 割り込み通知
		complete(&pStorage->compINT);
# endif
	}
	nIRQReturn = IRQ_HANDLED;
#else
   	{
		// CHANNEL Interrupt
		DWORD	dwIntSource;

		dwIntSource = ((dwINT >> 8) & 0x07);
		switch (dwIntSource) {
		default:
		case 0:	// NONE
		case 1:	// RXRDY and RX Line Status
		case 2:	// RXRDY Time-out
		case 3:	// TXRDY, THR or TSR empty
		case 4:	// MSR, RTS/CTS or DTR/DSR delta or Xoff/Xon det. or special char. detected.
		case 5:	// Reserved
			break;
		case 6:	// MPIO pin(s) ※CHANNEL #0
			{
				volatile BYTE	ucData;

				// 割り込み解除
				ucData = readb(&pREGs->R_MPIOLVL);	// Clear IRQ
				writeb(0x00, &pREGs->R_MPIOINT);	// MPIO割り込みを無効にする

#if 1
				// 割り込み通知
				up(&pStorage->semINT);
#else
				// 割り込み通知
				complete(&pStorage->compINT);
#endif
			}
			nIRQReturn = IRQ_HANDLED;
			break;
		case 7:	// TIMER Time-put ※CHANNEL #0 only
			break;
		}
	}
#endif
	
	return nIRQReturn;
}


//------------------------------------------------------------------------------
// 割り込みシグナル配信用カーネルスレッド
//------------------------------------------------------------------------------
static int rexpci60d_isr_thread(void *pParam)
{
	T_DEVICE_STORAGE	*pStorage = (T_DEVICE_STORAGE *)pParam;
	
	printk(KERN_NOTICE DRIVER_NAME ": #%d rexpci60d_isr_thread(%08lX) begin.\n", pStorage->nID, (unsigned long)pParam);
	
//	INIT_COMPLETION(&s_psDriver->sDevice[i].compINT);
//	while (!kthread_should_stop() && s_psDriver != NULL) {	// ドライバが生きている間は動作
	while (!kthread_should_stop() && !pStorage->bKillIntTask) {
		int					nSignal;
		struct task_struct	*pTask;
		
#if 1
		// 割り込み待ち
		if (down_interruptible(&pStorage->semINT)) {
			break;
		}
#else
		// 割り込み待ち
		wait_for_completion(&s_psDriver->sDevice[i].compINT);
		INIT_COMPLETION(&s_psDriver->sDevice[i].compINT);
#endif
		
		// リソースをロック
		down(&pStorage->semDevice);
		nSignal = pStorage->sSigInfo.nSignal;
		pTask = pStorage->sSigInfo.pTask;
		if (nSignal != 0) {
			// シグナルを送信する前にシグナルの登録を解除する(割り込み検出で自動解除)
			pStorage->sSigInfo.nSignal = 0;
			pStorage->sSigInfo.pTask = NULL;
			
			// 指定されたタスクへシグナルを送信
			send_sig(nSignal, pTask, 1);
		}
		
		// ioctlに通知
		if (pStorage->bWaitIoctl) {
			pStorage->bWaitIoctl = FALSE;
			pStorage->nIntStat = 1;
#if 1
			up(&pStorage->semIoctlINT);
#else
			complete(&pStorage->compIntWait);
#endif
		}
		
		// リソースロックを解除
		up(&pStorage->semDevice);
	}
	
	printk(KERN_NOTICE DRIVER_NAME ": #%d rexpci60d_isr_thread() end.\n", pStorage->nID);
	
	return 0;
}


//==============================================================================
// ドライバの登録と削除
//==============================================================================
#ifdef _USE_XR17C15X
static T_IDS_DRIVER_DATA initRegsREX_PCI64D;
static T_IDS_DRIVER_DATA initRegsREX_PCI60D;
static T_IDS_DRIVER_DATA initRegsREX_PCI70D;
#endif /* _USE_XR17C15X */
#ifdef _USE_XR17V35X
static T_IDS_DRIVER_DATA initRegsREX_PE64D;
static T_IDS_DRIVER_DATA initRegsREX_PE60D;
static T_IDS_DRIVER_DATA initRegsREX_PE70D;
#endif /* _USE_XR17V35X */

//------------------------------------------------------------------------------
// ドライバエントリ初期化処理
// static int rexpci60d_probe(struct pci_dev *link, const struct pci_device_id *id);
//   link     (IN)   pci_dev構造体のポインタ
//   id       (IN)   pci_device_id構造体のポインタ
//   <RETURN>       0:OK
//------------------------------------------------------------------------------
#ifdef _USE_SINGLE_MODULE
static int rexpci60d_probe(struct pci_dev *link, const struct pci_device_id *id)
#else
static int rexpci60d_probe(struct pci_dev *link, const struct pci_device_id *id, unsigned int **param)
#endif /* _USE_SINGLE_MODULE */
{
	unsigned long		ulBAR0_Base;	// BAR0 ベースアドレス
	unsigned long		ulBAR0_Len;		// BAR0 領域長
	void __iomem		*pIOBase;		// BAR0 のioremap()したアドレス
	T_XR17D154_DIO_REGS	*pREGs;			// DIO関連レジスタのアクセス用ポインタ
	T_IDS_DRIVER_DATA	*pIDS;			// ドライバ初期化データ
	
	struct device		*pDevice;		// クラスデバイス
	
	int					nDevNo;			// デバイス番号(DIPSWによる0～3の値)
	int					nReturn;		// int 戻り値ワーク
	
	int					i;
	
	printk(KERN_NOTICE DRIVER_NAME ": rexpci60d_probe(%08lX, %08lX)\n", (unsigned long)link, (unsigned long)id);
	
//	if (id->driver_data >= ARRAY_SIZE(rexPx6x70D_DIO_ids)) {
//		// driver_dataの値が不正
//		return -EINVAL;
//	};
#ifdef _USE_SINGLE_MODULE
	if (id->driver_data == 0) {
		// driver_dataの値が不正(driver_dataはT_IDS_DRIVER_DATAへのポインタの筈)
		return -EINVAL;
	}
#endif /* _USE_SINGLE_MODULE */
	// probeされるデバイスが、このドライバで面倒を見るものか確認する
	if (VENDOR_IO_RATOC != link->subsystem_vendor) {
		// RATOCボードじゃない
		return -EINVAL;
	}
	switch (link->subsystem_device) {
#ifdef _USE_SINGLE_MODULE
	case SUBSYS_ID_REX_PCI64D:	// REX-PCI64D
	case SUBSYS_ID_REX_PCI70D:	// REX-PCI70D
	case SUBSYS_ID_REX_PCI60D:	// REX-PCI60D
	case SUBSYS_ID_REX_PE64D:	// REX-PE64D
	case SUBSYS_ID_REX_PE70D:	// REX-PE70D
	case SUBSYS_ID_REX_PE60D:	// REX-PE60D
		break;
#else
#ifdef _USE_XR17C15X
	case SUBSYS_ID_REX_PCI64D:	// REX-PCI64D
		pIDS = &initRegsREX_PCI64D;
		break;
	case SUBSYS_ID_REX_PCI70D:	// REX-PCI70D
		pIDS = &initRegsREX_PCI70D;
		break;
	case SUBSYS_ID_REX_PCI60D:	// REX-PCI60D
		pIDS = &initRegsREX_PCI60D;
		break;
#endif /* _USE_XR17C15X */
#ifdef _USE_XR17V35X
	case SUBSYS_ID_REX_PE64D:	// REX-PE64D
		pIDS = &initRegsREX_PE64D;
		break;
	case SUBSYS_ID_REX_PE70D:	// REX-PE70D
		pIDS = &initRegsREX_PE70D;
		break;
	case SUBSYS_ID_REX_PE60D:	// REX-PE60D
		pIDS = &initRegsREX_PE60D;
		break;
#endif /* _USE_XR17V35X */
#endif /* _USE_SINGLE_MODULE */
	default:
		// このドライバでサポートするボードではない
		return -EINVAL;
	}
	
#ifdef _USE_SINGLE_MODULE
	nReturn = pci_enable_device(link);
	if (nReturn) {
		return nReturn;
	}
#endif /* _USE_SINGLE_MODULE */
	
	// ドライバ固有変数領域を確保(未確保の場合、つまり初回のみ)
	if (s_psDriver == NULL) {
		s_psDriver = (T_DRIVER_STORAGE *)kmalloc(sizeof(T_DRIVER_STORAGE), GFP_KERNEL);
		if (s_psDriver == NULL) {
			// "Driver storage can't allocated.\n"
			printk(KERN_NOTICE DRIVER_NAME ": Driver storage can't allocated.\n");
			return -ENOMEM;
		}
		
		// 使用カウンタをクリアw
		s_psDriver->nUsed = 0;
		
		// デバイス毎の領域をクリア
		for (i = 0; i < REX_DIO_MAX_BOARDS; i++) {
			sema_init(&s_psDriver->sDevice[i].semDevice, 1);	// デバイス排他制御用セマフォ
			s_psDriver->sDevice[i].bEnable = FALSE;
			
			s_psDriver->sDevice[i].pIDS = NULL;
			
			s_psDriver->sDevice[i].nSubSysID = -1;
			s_psDriver->sDevice[i].nID = -1;
			s_psDriver->sDevice[i].pDev = NULL;
			s_psDriver->sDevice[i].pDevice = NULL;
			
			s_psDriver->sDevice[i].nIRQ = -1;
			s_psDriver->sDevice[i].pBase = NULL;
			s_psDriver->sDevice[i].pREGs = NULL;
			
			s_psDriver->sDevice[i].uUserMask = 0x00;
			s_psDriver->sDevice[i].uUserMaskH = 0x00;
			
			s_psDriver->sDevice[i].sSigInfo.pFile = NULL;
			s_psDriver->sDevice[i].sSigInfo.nSignal = 0;
			s_psDriver->sDevice[i].sSigInfo.pTask = NULL;
			
			sema_init(&s_psDriver->sDevice[i].semINT, 0);
			s_psDriver->sDevice[i].pIntTask = NULL;
			s_psDriver->sDevice[i].bKillIntTask = FALSE;
			sema_init(&s_psDriver->sDevice[i].semIoctlINT, 0);
			s_psDriver->sDevice[i].nIntStat = 0;
			init_timer(&s_psDriver->sDevice[i].sIntTimer);
		}
	}
	
	// probeの時点でデバイスは確定しているので直ちにDIPSWを確認する(DIPSWの設定でnDevNoを決定)
	ulBAR0_Base = pci_resource_start(link, 0);	// BAR0のベースアドレス
	ulBAR0_Len = pci_resource_len(link, 0);		// BAR0の領域長
	pIOBase = ioremap_nocache(ulBAR0_Base, ulBAR0_Len);
	pREGs = (T_XR17D154_DIO_REGS *)(&((char *)pIOBase)[0x0080]);	// DIO関連レジスタの先頭アドレス
#ifdef _USE_SINGLE_MODULE
	pIDS = (T_IDS_DRIVER_DATA *)id->driver_data;
#endif /* _USE_SINGLE_MODULE */
	
	// DIPSWのIDを読み取りデバイスの番号(0～3)を決定する
	nDevNo = pIDS->pfnReadDipsw(pIDS, pIOBase);
	
	// リソースをロック
	down(&s_psDriver->sDevice[nDevNo].semDevice);
	
	// ドライバ変数領域を初期化
	if (s_psDriver->sDevice[nDevNo].bEnable) {
		// 既に同じIDのボードが設定済み
		printk(KERN_NOTICE DRIVER_NAME ": Duplicate DIPSW-ID(%d) found.\n", nDevNo);
		nReturn = -EINVAL;
		if(NULL != pIOBase)
		{
			iounmap(pIOBase);
			pIOBase = NULL;
		}
		goto err_out00;
	}

	s_psDriver->sDevice[nDevNo].bEnable   = TRUE;
	s_psDriver->sDevice[nDevNo].pIDS      = pIDS;
	
	s_psDriver->sDevice[nDevNo].nSubSysID = id->subdevice;
	s_psDriver->sDevice[nDevNo].nID       = nDevNo;
	s_psDriver->sDevice[nDevNo].pDev      = link;
	s_psDriver->sDevice[nDevNo].pDevice   = NULL;	// 後で設定
	
	s_psDriver->sDevice[nDevNo].nIRQ      = -1;		// 後で設定
	s_psDriver->sDevice[nDevNo].pBase     = pIOBase;	// 終了時にiounmap()するために保存
	s_psDriver->sDevice[nDevNo].pREGs     = pREGs;	// DIOレジスタアクセスのために保存
	
	s_psDriver->sDevice[nDevNo].uUserMask = (BYTE)(pIDS->wUserMask & 0x00FF);
	s_psDriver->sDevice[nDevNo].uUserMaskH = (BYTE)((pIDS->wUserMask >> 8) & 0x00FF);	// ※MPIO[15:8]は未使用
	
	s_psDriver->sDevice[nDevNo].sSigInfo.pFile = NULL;
	s_psDriver->sDevice[nDevNo].sSigInfo.nSignal = 0;
	s_psDriver->sDevice[nDevNo].sSigInfo.pTask = NULL;
	
	s_psDriver->sDevice[nDevNo].pIntTask = NULL;
	s_psDriver->sDevice[nDevNo].bKillIntTask = FALSE;
	s_psDriver->sDevice[nDevNo].nIntStat = 0;
	
	init_timer(&s_psDriver->sDevice[nDevNo].sIntTimer);
	
	// ドライバを初期化する
	if (s_psDriver->nUsed == 0) {
		// 初回のみ必要な初期化
		
		// キャラクタ・デバイス・ドライバ登録
		s_psDriver->nMajor = register_chrdev(REX_DIO_MAJOR_NUMBER, DRIVER_NAME, &dio_fops);
		if (s_psDriver->nMajor < 0) {
			// major番号登録エラー
			printk(KERN_NOTICE DRIVER_NAME ": register_chrdev() failed.\n");
			nReturn = s_psDriver->nMajor;
			goto err_out0;
		}
		
		// クラス生成 (/sys/class/<DRIVER_NAME>)
		s_psDriver->psClass = class_create(THIS_MODULE, DRIVER_NAME);
		if (IS_ERR(s_psDriver->psClass)) {
			// クラス生成エラー
			printk(KERN_NOTICE DRIVER_NAME ": class_create() failed.\n");
			nReturn = PTR_ERR(s_psDriver->psClass);
			goto err_out1;
		}
	}
	
	// クラスデバイス生成 (/sys/class/<DRIVER_NAME>/<device name> ==> udev:/dev/<device name>)
	pDevice = device_create(s_psDriver->psClass, NULL, MKDEV(s_psDriver->nMajor, nDevNo), NULL, s_szClassDeviceName[nDevNo]);
	if (IS_ERR(pDevice)) {
		// クラスデバイス生成エラー
		printk(KERN_NOTICE DRIVER_NAME ": device_create() failed.\n");
		nReturn = PTR_ERR(pDevice);
		goto err_out2;
	}
	s_psDriver->sDevice[nDevNo].pDevice = pDevice;
#ifdef _USE_SINGLE_MODULE
	pci_set_drvdata(link, &s_psDriver->sDevice[nDevNo]);
#else
	*param = (unsigned int *)&s_psDriver->sDevice[nDevNo];
#endif /* _USE_SINGLE_MODULE */
	
	// 割り込み番号の取得
#if 1
	s_psDriver->sDevice[nDevNo].nIRQ = link->irq;	// PCI IRQ
#else
	nReturn = pci_read_config_byte(link, PCI_INTERRUPT_LINE, &s_psDriver->sDevice[nDevNo].nIRQ);
	if (nReturn != 0) {
		// 割り込み番号取得失敗
		printk(KERN_NOTICE DRIVER_NAME ": pci_read_config_byte() failed.\n");
		goto err_out3;
	}
#endif
	
	// 割り込みハンドラを登録
	nReturn = request_irq(s_psDriver->sDevice[nDevNo].nIRQ, rexpci60d_isr, IRQF_DISABLED | IRQF_SHARED, DRIVER_NAME, &s_psDriver->sDevice[nDevNo]);
	if (nReturn != 0) {
		printk(KERN_NOTICE DRIVER_NAME ": request_irq() failed.\n");
		goto err_out3;
	}
	printk(KERN_NOTICE DRIVER_NAME ": #%d IRQ %d driver initialzed.\n", nDevNo, s_psDriver->sDevice[nDevNo].nIRQ);
	
	// ハードウェア初期化
	pIDS->pfnInitHW(pIDS, pIOBase);
	
	// 使用カウンタを加算
	s_psDriver->nUsed++;
	
	// リソースをロック解除
	up(&s_psDriver->sDevice[nDevNo].semDevice);
	
	// デバイス初回起床
//	device_init_wakeup(&pDev->dev, 1);
	
	// 割り込み処理スレッドを起動
	s_psDriver->sDevice[nDevNo].pIntTask = kthread_run(rexpci60d_isr_thread, &s_psDriver->sDevice[nDevNo], "rexpci60d_isr_thread");
	if (IS_ERR(s_psDriver->sDevice[nDevNo].pIntTask)) {
		printk(KERN_NOTICE DRIVER_NAME ": kthread_run() failed.\n");
		nReturn = PTR_ERR(s_psDriver->sDevice[nDevNo].pIntTask);
		s_psDriver->sDevice[nDevNo].pIntTask = NULL;
		goto err_out4;
	}
	
	return 0;	// 正常終了
	
// request_irq()以降での異状処理エントリ
err_out4:
	free_irq(s_psDriver->sDevice[nDevNo].nIRQ, &s_psDriver->sDevice[nDevNo]);
	
// device_create()以降での異状処理エントリ
err_out3:
	device_destroy(s_psDriver->psClass, MKDEV(s_psDriver->nMajor, nDevNo));
	
// class_create()以降での異状処理エントリ
err_out2:
	class_destroy(s_psDriver->psClass);

// register_chrdev()以降での異状処理エントリ
err_out1:
	if (REX_DIO_MAJOR_NUMBER == 0) {
		unregister_chrdev(s_psDriver->nMajor, DRIVER_NAME);
	}
	else {
		unregister_chrdev(REX_DIO_MAJOR_NUMBER, DRIVER_NAME);
	}
	
// kmalloc() 以降での異状処理エントリ
err_out0:
	if (s_psDriver->sDevice[nDevNo].pBase != NULL) {
		iounmap(s_psDriver->sDevice[nDevNo].pBase);
		s_psDriver->sDevice[nDevNo].pBase = NULL;
	}
	s_psDriver->sDevice[nDevNo].bEnable = FALSE;
	
// ID重複 検出での異常処理エントリ
err_out00:
	// リソースをロック解除
	up(&s_psDriver->sDevice[nDevNo].semDevice);
	
	if (s_psDriver->nUsed == 0) {
		kfree(s_psDriver);
		s_psDriver = NULL;
	}
	
#ifdef _USE_SINGLE_MODULE
	pci_disable_device(link);
#else
	*param = NULL;
#endif /* _USE_SINGLE_MODULE */
	
	return nReturn;
}

//------------------------------------------------------------------------------
// ドライバエントリ終了処理
// static void rexpci60d_remove(struct pci_dev *link);
//   link     (IN)   pci_dev構造体のポインタ
//------------------------------------------------------------------------------
#ifdef _USE_SINGLE_MODULE
static void rexpci60d_remove(struct pci_dev *link)
#else
static void rexpci60d_remove(struct pci_dev *link, unsigned int **param)
#endif /* _USE_SINGLE_MODULE */
{
#ifdef _USE_SINGLE_MODULE
	T_DEVICE_STORAGE	*pStorage = pci_get_drvdata(link);
#else
	T_DEVICE_STORAGE	*pStorage = (T_DEVICE_STORAGE *)*param;
#endif /* _USE_SINGLE_MODULE */
	int					nDevNo;
	
	printk(KERN_NOTICE DRIVER_NAME ": rexpci60d_remove(%08lX)\n", (unsigned long)link);
	
	nDevNo = pStorage->nID;
	
	// 割り込み処理タスクを停止
	if (s_psDriver->sDevice[nDevNo].pIntTask != NULL) {
		// 割り込み処理タスクに対し停止要求
		s_psDriver->sDevice[nDevNo].bKillIntTask = TRUE;
		up(&s_psDriver->sDevice[nDevNo].semINT);	// 割り込みイベント待ちを解除するためのup
		
		// 割り込み処理タスクの停止を待つ
		kthread_stop(s_psDriver->sDevice[nDevNo].pIntTask);
		s_psDriver->sDevice[nDevNo].pIntTask = NULL;
	}
	
	// ハードウェア後始末
	s_psDriver->sDevice[nDevNo].pIDS->pfnDeinitHW(s_psDriver->sDevice[nDevNo].pIDS, s_psDriver->sDevice[nDevNo].pBase);
	
	// リソースをロック
//	down(&s_psDriver->sDevice[nDevNo].semDevice);
	
	// 割り込みハンドラ解放
	free_irq(s_psDriver->sDevice[nDevNo].nIRQ, &s_psDriver->sDevice[nDevNo]);
	
	// デバイス削除
	device_destroy(s_psDriver->psClass, MKDEV(s_psDriver->nMajor, nDevNo));
	
	if (s_psDriver->nUsed == 1) {
		// クラス削除
		class_destroy(s_psDriver->psClass);
		
		// キャラクタ・デバイス・ドライバ登録解除
		if (REX_DIO_MAJOR_NUMBER == 0) {
			unregister_chrdev(s_psDriver->nMajor, DRIVER_NAME);
		}
		else {
			unregister_chrdev(REX_DIO_MAJOR_NUMBER, DRIVER_NAME);
		}
	}
	s_psDriver->nUsed--;
	
	// I/O mapping 解除
	if (s_psDriver->sDevice[nDevNo].pBase != NULL) {
		iounmap(s_psDriver->sDevice[nDevNo].pBase);
		s_psDriver->sDevice[nDevNo].pBase = NULL;
	}
	s_psDriver->sDevice[nDevNo].bEnable = FALSE;
	
	// リソースをロック解除
//	up(&s_psDriver->sDevice[nDevNo].semDevice);
	
	if (s_psDriver->nUsed == 0) {
		// ドライバ固有変数領域解放
		kfree(s_psDriver);
		s_psDriver = NULL;
	}
	
#ifdef _USE_SINGLE_MODULE
	pci_set_drvdata(link, NULL);
#else
	*param = NULL;
#endif /* _USE_SINGLE_MODULE */
	
#ifdef _USE_SINGLE_MODULE
	pci_disable_device(link);
#endif /* _USE_SINGLE_MODULE */
	printk(KERN_NOTICE DRIVER_NAME ": #%d DIO Driver disabled\n", nDevNo);
}

#if 0
//------------------------------------------------------------------------------
// ドライバSUSPEND処理
//------------------------------------------------------------------------------
static int rexpci60d_suspend(struct pci_dev *link, u32 state)
{
	printk(KERN_NOTICE DRIVER_NAME ": rexpci60d_suspend(%08lX, %08lX)\n", (unsigned long)link, (unsigned long)state);
	return 0;
}

//------------------------------------------------------------------------------
// ドライバRESUME処理
//------------------------------------------------------------------------------
static int rexpci60d_resume(struct pci_dev *link)
{
	printk(KERN_NOTICE DRIVER_NAME ": rexpci60d_resume(%08lX)\n", (unsigned long)link);
	return 0;
}
#endif

//==============================================================================
// ハードウェア固有処理
// ※driver_data経由でprobe()/remove()から呼び出される
//==============================================================================

#ifdef _USE_XR17C15X
//------------------------------------------------------------------------------
// XR17D15x用ハードウェア初期化処理
// static int initXR17D15x(T_IDS_DRIVER_DATA *pIDS, void __iomem *pBase)
//   pIDS    : T_IDS_DRIVER_DATA 情報のアドレス
//   pBase   : XR17D15x ハードウェアレジスタのベースアドレス(ioremap()されたもの)
//   <RETURN>: TRUE=OK, FALSE=NG
// ※driver_data経由でprobe処理から呼び出される
//------------------------------------------------------------------------------
static int initXR17D15x(T_IDS_DRIVER_DATA *pIDS, void __iomem *pBase)
{
	T_XR17D154_DIO_REGS	*pREGs;
	
	DEBUG_PRINTK(KERN_NOTICE DRIVER_NAME ": initXR17D15x(%08lX, %08lX)\n", (unsigned long)pIDS, (unsigned long)pBase);
	
	pREGs = (T_XR17D154_DIO_REGS *)&(((char *)pBase)[0x0080]);	// DIO関連レジスタの先頭アドレス
	
	// ハードウェアレジスタを初期状態に設定
	writeb(pIDS->sInitREG.MPIOINT, &pREGs->R_MPIOINT);
	writeb(pIDS->sInitREG.MPIOLVL, &pREGs->R_MPIOLVL);
	writeb(pIDS->sInitREG.MPIO3T,  &pREGs->R_MPIO3T);
	writeb(pIDS->sInitREG.MPIOINV, &pREGs->R_MPIOINV);
	writeb(pIDS->sInitREG.MPIOSEL, &pREGs->R_MPIOSEL);
	
	return TRUE;
}
#endif /* _USE_XR17C15X */

#ifdef _USE_XR17V35X
//------------------------------------------------------------------------------
// XR17D35x用ハードウェア初期化処理
// static int initXR17D15x(T_IDS_DRIVER_DATA *pIDS, void __iomem *pBase)
//   pIDS    : T_IDS_DRIVER_DATA 情報のアドレス
//   pBase   : XR17D15x ハードウェアレジスタのベースアドレス(ioremap()されたもの)
//   <RETURN>: TRUE=OK, FALSE=NG
// ※driver_data経由でprobe処理から呼び出される
//------------------------------------------------------------------------------
static int initXR17D35x(T_IDS_DRIVER_DATA *pIDS, void __iomem *pBase)
{
	T_XR17D354_DIO_REGS	*pREGs;
	
	DEBUG_PRINTK(KERN_NOTICE DRIVER_NAME ": initXR17D35x(%08lX, %08lX)\n", (unsigned long)pIDS, (unsigned long)pBase);
	
	pREGs = (T_XR17D354_DIO_REGS *)&(((char *)pBase)[0x0080]);	// DIO関連レジスタの先頭アドレス
	
	// ハードウェアレジスタを初期状態に設定
	// MPIO[7:0]
	writeb(pIDS->sInitREG.MPIOINT, &pREGs->R_MPIOINT);
	writeb(pIDS->sInitREG.MPIOLVL, &pREGs->R_MPIOLVL);
	writeb(pIDS->sInitREG.MPIO3T,  &pREGs->R_MPIO3T);
	writeb(pIDS->sInitREG.MPIOINV, &pREGs->R_MPIOINV);
	writeb(pIDS->sInitREG.MPIOSEL, &pREGs->R_MPIOSEL);
	writeb(pIDS->sInitREG.MPIOOD,  &pREGs->R_MPIOOD);
	// MPIO[15:8]
	writeb(pIDS->sInitREGH.MPIOINT, &pREGs->R_MPIOINT_H);
	writeb(pIDS->sInitREGH.MPIOLVL, &pREGs->R_MPIOLVL_H);
	writeb(pIDS->sInitREGH.MPIO3T,  &pREGs->R_MPIO3T_H);
	writeb(pIDS->sInitREGH.MPIOINV, &pREGs->R_MPIOINV_H);
	writeb(pIDS->sInitREGH.MPIOSEL, &pREGs->R_MPIOSEL_H);
	writeb(pIDS->sInitREGH.MPIOOD,  &pREGs->R_MPIOOD_H);
	
	return TRUE;
}
#endif /* _USE_XR17V35X */

//------------------------------------------------------------------------------
// REX-PCI64D用DIPSW読み出し処理
// static int readDIPSW_REX_PCI64D((T_IDS_DRIVER_DATA *pIDS, void __iomem *pBase)
//   pIDS    : T_IDS_DRIVER_DATA 情報のアドレス
//   pBase   : XR17D15x ハードウェアレジスタのベースアドレス(ioremap()されたもの)
//   <RETURN>: DIPSWの値(0～3)
// ※driver_data経由でprobe処理から呼び出される
//------------------------------------------------------------------------------
static int readDIPSW_REX_PCI64D(T_IDS_DRIVER_DATA *pIDS, void __iomem *pBase)
{
	T_XR17D154_DIO_REGS	*pREGs;
	int					nID;
	BYTE				uDATA;
	
	DEBUG_PRINTK(KERN_NOTICE DRIVER_NAME ": readDIPSW_REX_PCI64D(%08lX, %08lX)\n", (unsigned long)pIDS, (unsigned long)pBase);
	
	pREGs = (T_XR17D154_DIO_REGS *)&(((char *)pBase)[0x0080]);	// DIO関連レジスタの先頭アドレス
	uDATA = readb(&pREGs->R_MPIOLVL) ^ readb(&pREGs->R_MPIOINV);
	nID = (~(uDATA >> 6)) & 0x03;	// bit6&7がID(負論理)

	return nID;
}

//------------------------------------------------------------------------------
// REX-PCI60/70D用DIPSW読み出し処理
// static int readDIPSW_REX_PCI60D((T_IDS_DRIVER_DATA *pIDS, void __iomem *pBase)
//   pIDS    : T_IDS_DRIVER_DATA 情報のアドレス
//   pBase   : XR17D15x ハードウェアレジスタのベースアドレス(ioremap()されたもの)
//   <RETURN>: DIPSWの値(0～3)
// ※driver_data経由でprobe処理から呼び出される
//------------------------------------------------------------------------------
static int readDIPSW_REX_PCI60D(T_IDS_DRIVER_DATA *pIDS, void __iomem *pBase)
{
	T_XR17D154_DIO_REGS	*pREGs;
	int					nID;
	BYTE				uDATA;
	int					nConvert[4] = {0x00, 0x02, 0x01, 0x03};
	
	DEBUG_PRINTK(KERN_NOTICE DRIVER_NAME ": readDIPSW_REX_PCI60D(%08lX, %08lX)\n", (unsigned long)pIDS, (unsigned long)pBase);
	
	pREGs = (T_XR17D154_DIO_REGS *)&(((char *)pBase)[0x0080]);	// DIO関連レジスタの先頭アドレス
	uDATA = readb(&pREGs->R_MPIOLVL) ^ readb(&pREGs->R_MPIOINV);
	nID = (~(uDATA >> 6)) & 0x03;	// bit6&7がID(負論理)
	nID = nConvert[nID];			// bit6&7の結線が逆のため変換
	
	return nID;
}

//==============================================================================
// ドライバ登録関連
//==============================================================================

//------------------------------------------------------------------------------
// ドライバ初期化情報
//------------------------------------------------------------------------------

#ifdef _USE_XR17C15X
static T_IDS_DRIVER_DATA	initRegsREX_PCI64D = {
	.sInitREG = {
		0x00,	// MPIOINT - 割り込みdisabled
		0x00,	// MPIOLVL - ALL LOW
		0x00,	// MPIO3T  - ALL ACTIVE (not tri-state)
		0x00,	// MPIOINV - ALL NORMAL (not invert)
		0xFF,	// MPIOSEL - ALL INPUT
		0x00,	// MPIOOD  - ※未使用
	},
	.wUserMask = 0x003F,	// MPIO0～5の6ビットを使用可能, MPIO6～MPIO7はDIPSW用なので使用禁止
	.pfnInitHW = initXR17D15x,				// ハードウェア初期化(XR17D15x)
	.pfnDeinitHW = initXR17D15x,			// ハードウェア後始末(XR17D15x)
	.pfnReadDipsw = readDIPSW_REX_PCI64D,	// DIPSW読み出し処理はPCI64D用
};

static T_IDS_DRIVER_DATA	initRegsREX_PCI60D = {
	.sInitREG = {
		0x00,	// MPIOINT - 割り込みdisabled
		0x00,	// MPIOLVL - ALL LOW
		0x00,	// MPIO3T  - ALL ACTIVE (not tri-state)
		0x00,	// MPIOINV - ALL NORMAL (not invert)
		0xFF,	// MPIOSEL - ALL INPUT
		0x00,	// MPIOOD  - ※未使用
	},
	.wUserMask = 0x003F,	// MPIO0～5の6ビットを使用可能, MPIO6～MPIO7はDIPSW用なので使用禁止
	.pfnInitHW = initXR17D15x,				// ハードウェア初期化(XR17D15x)
	.pfnDeinitHW = initXR17D15x,			// ハードウェア後始末(XR17D15x)
	.pfnReadDipsw = readDIPSW_REX_PCI60D,	// DIPSW読み出し処理はPCI60/70D用
};

static T_IDS_DRIVER_DATA	initRegsREX_PCI70D = {
	.sInitREG = {
		0x00,	// MPIOINT - 割り込みdisabled
		0x00,	// MPIOLVL - ALL LOW
		0x00,	// MPIO3T  - ALL ACTIVE (not tri-state)
		0x00,	// MPIOINV - ALL NORMAL (not invert)
		0xFF,	// MPIOSEL - ALL INPUT
		0x00,	// MPIOOD  - ※未使用
	},
	.wUserMask = 0x003F,	// MPIO0～5の6ビットを使用可能, MPIO6～MPIO7はDIPSW用なので使用禁止
	.pfnInitHW = initXR17D15x,				// ハードウェア初期化(XR17D15x)
	.pfnDeinitHW = initXR17D15x,			// ハードウェア後始末(XR17D15x)
	.pfnReadDipsw = readDIPSW_REX_PCI60D,	// DIPSW読み出し処理はPCI60/70D用
};
#endif /* _USE_XR17C15X */

#ifdef _USE_XR17V35X
static T_IDS_DRIVER_DATA	initRegsREX_PE64D = {
	.sInitREG = {
		0x00,	// MPIOINT - 割り込みdisabled
		0x00,	// MPIOLVL - ALL LOW
		0x00,	// MPIO3T  - ALL ACTIVE (not tri-state)
		0x00,	// MPIOINV - ALL NORMAL (not invert)
		0xFF,	// MPIOSEL - ALL INPUT
		0x00,	// MPIOOD  - ALL NORMAL
	},
	.sInitREGH = {
		0x00,	// MPIOINT - 割り込みdisabled
		0x00,	// MPIOLVL - ALL LOW
		0x00,	// MPIO3T  - ALL ACTIVE (not tri-state)
		0x00,	// MPIOINV - ALL NORMAL (not invert)
		0xFF,	// MPIOSEL - ALL INPUT
		0x00,	// MPIOOD  - ALL NORMAL
	},
	.wUserMask = 0x003F,	// MPIO0～5の6ビットを使用可能, MPIO6～MPIO7はDIPSW用なので使用禁止、MPIO8～15は利用不可
	.pfnInitHW = initXR17D35x,				// ハードウェア初期化(XR17D35x)
	.pfnDeinitHW = initXR17D35x,			// ハードウェア後始末(XR17D35x)
	.pfnReadDipsw = readDIPSW_REX_PCI64D,	// DIPSW読み出し処理はPCI64D用
};

static T_IDS_DRIVER_DATA	initRegsREX_PE60D = {
	.sInitREG = {
		0x00,	// MPIOINT - 割り込みdisabled
		0x00,	// MPIOLVL - ALL LOW
		0x00,	// MPIO3T  - ALL ACTIVE (not tri-state)
		0x00,	// MPIOINV - ALL NORMAL (not invert)
		0xFF,	// MPIOSEL - ALL INPUT
		0x00,	// MPIOOD  - ALL NORMAL
	},
	.sInitREGH = {
		0x00,	// MPIOINT - 割り込みdisabled
		0x00,	// MPIOLVL - ALL LOW
		0x00,	// MPIO3T  - ALL ACTIVE (not tri-state)
		0x00,	// MPIOINV - ALL NORMAL (not invert)
		0xFF,	// MPIOSEL - ALL INPUT
		0x00,	// MPIOOD  - ALL NORMAL
	},
	.wUserMask = 0x003F,	// MPIO0～5の6ビットを使用可能, MPIO6～MPIO7はDIPSW用なので使用禁止、MPIO8～15は利用不可
	.pfnInitHW = initXR17D35x,				// ハードウェア初期化(XR17D35x)
	.pfnDeinitHW = initXR17D35x,			// ハードウェア後始末(XR17D35x)
	.pfnReadDipsw = readDIPSW_REX_PCI60D,	// DIPSW読み出し処理はPCI60/70D用
};

static T_IDS_DRIVER_DATA	initRegsREX_PE70D = {
	.sInitREG = {
		0x00,	// MPIOINT - 割り込みdisabled
		0x00,	// MPIOLVL - ALL LOW
		0x00,	// MPIO3T  - ALL ACTIVE (not tri-state)
		0x00,	// MPIOINV - ALL NORMAL (not invert)
		0xFF,	// MPIOSEL - ALL INPUT
		0x00,	// MPIOOD  - ※未使用
	},
	.sInitREGH = {
		0x00,	// MPIOINT - 割り込みdisabled
		0x00,	// MPIOLVL - ALL LOW
		0x00,	// MPIO3T  - ALL ACTIVE (not tri-state)
		0x00,	// MPIOINV - ALL NORMAL (not invert)
		0xFF,	// MPIOSEL - ALL INPUT
		0x00,	// MPIOOD  - ※未使用
	},
	.wUserMask = 0x003F,	// MPIO0～5の6ビットを使用可能, MPIO6～MPIO7はDIPSW用なので使用禁止、MPIO8～15は利用不可
	.pfnInitHW = initXR17D35x,				// ハードウェア初期化(XR17D35x)
	.pfnDeinitHW = initXR17D35x,			// ハードウェア後始末(XR17D35x)
	.pfnReadDipsw = readDIPSW_REX_PCI60D,	// DIPSW読み出し処理はPCI60/70D用
};
#endif /* _USE_XR17V35X */

#ifdef _USE_SINGLE_MODULE
//------------------------------------------------------------------------------
// PCI DEVICE ID 情報テーブル(pci_device_id 構造体)
// ※このドライバがサポートするデバイスの情報
//------------------------------------------------------------------------------
static struct pci_device_id rexPx6x70D_DIO_ids[] = {
#ifdef _USE_XR17C15X
	{	// REX-PCI64D PCI 4port RS-232C
		.vendor = VENDOR_ID_EXAR,
		.device = DEVICE_ID_XR17V154,
		.subvendor = VENDOR_IO_RATOC,
		.subdevice = SUBSYS_ID_REX_PCI64D,
		.driver_data = (kernel_ulong_t)&initRegsREX_PCI64D,
	},
	{	// REX-PCI60D PCI 2port RS-232C
		.vendor = VENDOR_ID_EXAR,
		.device = DEVICE_ID_XR17V152,
		.subvendor = VENDOR_IO_RATOC,
		.subdevice = SUBSYS_ID_REX_PCI60D,
		.driver_data = (kernel_ulong_t)&initRegsREX_PCI60D,
	},
	{	// REX-PCI70D PCI 2port RS-422/485
		.vendor = VENDOR_ID_EXAR,
		.device = DEVICE_ID_XR17V152,
		.subvendor = VENDOR_IO_RATOC,
		.subdevice = SUBSYS_ID_REX_PCI70D,
		.driver_data = (kernel_ulong_t)&initRegsREX_PCI70D,
	},
#endif /* _USE_XR17C15X */
#ifdef _USE_XR17V35X
	{	// REX-PE64D PCIe 4port RS-232C
		.vendor = VENDOR_ID_EXAR,
		.device = DEVICE_ID_XR17V354,
		.subvendor = VENDOR_IO_RATOC,
		.subdevice = SUBSYS_ID_REX_PE64D,
		.driver_data = (kernel_ulong_t)&initRegsREX_PE64D,
	},
	{	// REX-PE60D PCIe 2port RS-232C
		.vendor = VENDOR_ID_EXAR,
		.device = DEVICE_ID_XR17V352,
		.subvendor = VENDOR_IO_RATOC,
		.subdevice = SUBSYS_ID_REX_PE60D,
		.driver_data = (kernel_ulong_t)&initRegsREX_PE60D,
	},
	{	// REX-PE70D PCIe 2port RS-422/485
		.vendor = VENDOR_ID_EXAR,
		.device = DEVICE_ID_XR17V352,
		.subvendor = VENDOR_IO_RATOC,
		.subdevice = SUBSYS_ID_REX_PE70D,
		.driver_data = (kernel_ulong_t)&initRegsREX_PE70D,
	},
#endif /* _USE_XR17V35X */
	{ 0, }
};

//==============================================================================
// 単体DIOカーネルモジュール用
//==============================================================================
//------------------------------------------------------------------------------
// PCI DRIVER 情報テーブル(pci_driver 構造体)
//------------------------------------------------------------------------------

static struct pci_driver rexPx6x70D_DIO_Driver = {
	.name = DRIVER_NAME,
	.id_table = rexPx6x70D_DIO_ids,
	.probe = rexpci60d_probe,
	.remove = rexpci60d_remove,
#if 0
	.suspend = rexpci60d_suspend,
	.resume = rexpci60d_resume,
#endif
};

//------------------------------------------------------------------------------
// ローダブルモジュール初期化処理
// static int __init rexpci60d_init_module(void);
//   <RETURN>       0:OK, !0:NG
//------------------------------------------------------------------------------
static int __init rexpci60d_init_module(void)
{
	printk(KERN_NOTICE DRIVER_NAME ": %s Digital I/O Driver V.%s\n", VER_MSG, DRIVER_VERSION);
#ifdef USE_DEBUG_PROC_REXDIO
	// /proc/rexdio エントリ生成
	rexdio_create_proc();
#endif
	return pci_register_driver(&rexPx6x70D_DIO_Driver);
}

//------------------------------------------------------------------------------
// ローダブルモジュール終了処理
// static void __exit rexpci60d_exit_module(void);
//------------------------------------------------------------------------------
static void __exit rexpci60d_exit_module(void)
{
#ifdef USE_DEBUG_PROC_REXDIO
	// /proc/rexdio エントリ削除
	rexdio_remove_proc();
#endif
	pci_unregister_driver(&rexPx6x70D_DIO_Driver);
}

//------------------------------------------------------------------------------
// module エントリ
//------------------------------------------------------------------------------
module_init(rexpci60d_init_module);
module_exit(rexpci60d_exit_module);

//------------------------------------------------------------------------------
// module 情報
//------------------------------------------------------------------------------
MODULE_DESCRIPTION("REX-PCI60D Digital I/O Driver");
MODULE_DEVICE_TABLE(pci, rexPx6x70D_DIO_ids);

MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("RATOC SYSTEMS INC.");
MODULE_VERSION(DRIVER_VERSION);

#else
//==============================================================================
// シリアルドライバに統合
//==============================================================================

//------------------------------------------------------------------------------
// ローダブルモジュール初期化処理
// シリアルドライバから呼び出す
//------------------------------------------------------------------------------
int rexpci60d_init_module_wrapper(void)
{
	printk(KERN_NOTICE DRIVER_NAME ": %s Digital I/O Driver V.%s\n", VER_MSG, DRIVER_VERSION);

	return 0;
}

//------------------------------------------------------------------------------
// ローダブルモジュール終了処理
// シリアルドライバから呼び出す
//------------------------------------------------------------------------------
void rexpci60d_exit_module_wrapper(void)
{
}

//------------------------------------------------------------------------------
// ドライバエントリ初期化処理
// static int rexpci60d_probe(struct pci_dev *link, const struct pci_device_id *id, unsigned int **param);
//   dev      (IN)   pci_dev構造体のポインタ
//   ent      (IN)   pci_device_id構造体のポインタ
//   **param  (OUT)  pci_set_drvdataに登録するデータへのポインタ
//   <RETURN>       0:OK
//------------------------------------------------------------------------------
int rexpci60d_probe_wrapper(struct pci_dev *dev, const struct pci_device_id *ent, unsigned int **param)
{
	int ret;

	*param = NULL;
	ret = rexpci60d_probe(dev, ent, param);

	return ret;
}

//------------------------------------------------------------------------------
// ドライバエントリ終了処理
// static void rexpci60d_remove(struct pci_dev *link, unsigned int **param);
//   link     (IN)       pci_dev構造体のポインタ
//   **param  (IN/OUT)   pci_set_drvdataに登録したデータへのポインタ
//------------------------------------------------------------------------------
void rexpci60d_remove_wrapper(struct pci_dev *dev, unsigned int **param)
{
	if(NULL != *param)
	{
		rexpci60d_remove(dev, param);
	}
}

#endif /* _USE_SINGLE_MODULE */

