//==============================================================================
// rexpx6x70d_drv.h
//   REX-PE60D/64D/70D, REX-PCI60D/64D/70D DIOドライバ 定義ファイル
//
//  Copyright (C) 2012 RATOC Systems, Inc.
//  Dual licensed under the BSD (BSD-LICENSE.txt)
//  and GPL (GPL-LICENSE.txt) licenses.
//
//==============================================================================

#ifndef _REXPX6X70D_DRV_H
#define _REXPX6X70D_DRV_H
//==============================================================================
// ioctl() パラメタ構造体定義
//==============================================================================

//------------------------------------------------------------------------------
// IOCTL_DIO_READ_PORT/IOCTL_DIO_WRITE_PORT パラメタ構造体
//------------------------------------------------------------------------------
typedef struct T_IOPORT_UCHAR {
	unsigned int	Offset;		// XR17V532レジスタアドレス(0x80～0xFFのみ有効)
	unsigned char	Value;		// 値(1バイトアクセス)
} IOPORT_UCHAR, *PIOPORT_UCHAR;


//------------------------------------------------------------------------------
// REX_DIO_IOCG_READ パラメタ構造体
//------------------------------------------------------------------------------
typedef struct T_IOCTL_DIO_READ {
	unsigned char	ucData;			// 入力データ
} T_IOCTL_DIO_READ, *PT_IOCTL_DIO_READ;

//------------------------------------------------------------------------------
// REX_DIO_IOCS_WRITE パラメタ構造体
//------------------------------------------------------------------------------
typedef struct T_IOCTL_DIO_WRITE {
	unsigned char	ucData;			// 出力データ
} T_IOCTL_DIO_WRITE, *PT_IOCTL_DIO_WRITE;

//------------------------------------------------------------------------------
// REX_DIO_IOCS_INTSIG_ENABLE パラメタ構造体
//------------------------------------------------------------------------------
typedef struct T_IOCTL_DIO_INT_ENABLE {
	unsigned char	ucIntBit;		// 割り込みを有効にするビットマスク(1:Enable, 0:Disable)
	int				nSignalNo;		// 割り込み通知用シグナル番号
} T_IOCTL_DIO_INT_ENABLE, *PT_IOCTL_DIO_INT_ENABLE;

//------------------------------------------------------------------------------
// REX_DIO_IOCS_INTSIG_DISABLE パラメタ構造体
//------------------------------------------------------------------------------
typedef struct T_IOCTL_DIO_INT_DISABLE {
	int				nSignalNo;		// 割り込み通知を解除するシグナル番号(dummy)
} T_IOCTL_DIO_INT_DISABLE, *PT_IOCTL_DIO_INT_DISABLE;

//------------------------------------------------------------------------------
// REX_DIO_IOCS_DIRECTION パラメタ構造体
//------------------------------------------------------------------------------
typedef struct T_IOCTL_DIO_DIRECTION {
	unsigned char	ucDirection;	// I/O方向設定値(1:input, 0:output)
} T_IOCTL_DIO_DIRECTION, *PT_IOCTL_DIO_DIRECTION;

//------------------------------------------------------------------------------
// REX_DIO_IOCS_INVERT パラメタ構造体
//------------------------------------------------------------------------------
typedef struct T_IOCTL_DIO_INVERT {
	unsigned char	ucInvert;		// INVERT設定データ(1:invert, 0:not invert)
} T_IOCTL_DIO_INVERT, *PT_IOCTL_DIO_INVERT;

//------------------------------------------------------------------------------
// REX_DIO_IOCS_TRISTATE パラメタ構造体
//------------------------------------------------------------------------------
typedef struct T_IOCTL_DIO_TRISTATE {
	unsigned char	ucTriState;		// TRI-STATE設定データ(1:tri-state, 0:normal)
} T_IOCTL_DIO_TRISTATE, *PT_IOCTL_DIO_TRISTATE;


//==============================================================================
// ioctl() 機能コード等の定義
//==============================================================================
#define REX_DIO_IOC_MAGIC	'r'
// S(Set)      ポインタで「設定」
// T(Tell)     引数の値で直接「通知」
// G(Get)      ポインタで応答を「貰う」
// Q(Query)    「問い合わせ」の応答は戻り値
// X(eXchange) GとSをatomicに「交換」
// H(sHift)    TとQをatomicに「シフト」

// PORTからの読み出し
// ※オフセット0x80～0xFFの任意のレジスタから値を読み出します。
#define IOCTL_DIO_READ_PORT			_IOWR(REX_DIO_IOC_MAGIC, 0, IOPORT_UCHAR)

// PORTへの書込み
// ※オフセット0x80～0xFFの任意のレジスタに値を書き込みます。
#define IOCTL_DIO_WRITE_PORT		_IOW(REX_DIO_IOC_MAGIC, 1, IOPORT_UCHAR)

// 割り込み待ち開始
// ※DIO割り込みが発生するまで処理をブロックします。
#define IOCTL_DIO_START_INTERRUPT	_IO(REX_DIO_IOC_MAGIC, 2)

// 割り込み待ち終了
// ※IOCTL_DIO_START_INTERRUPTの動作を解除します。
#define IOCTL_DIO_STOP_INTERRUPT	_IO(REX_DIO_IOC_MAGIC, 3)

// 入力ポートからの読み出し
// ※MPIOLVLレジスタのデータを読み出します。
#define REX_DIO_IOCG_READ			_IOR(REX_DIO_IOC_MAGIC, 4, T_IOCTL_DIO_READ)

// 出力ポートへの書込み
// ※MPIOLVLレジスタへデータを書き込みます。
#define REX_DIO_IOCS_WRITE			_IOW(REX_DIO_IOC_MAGIC, 5, T_IOCTL_DIO_WRITE)

// 割り込みシグナル設定
// ※DIO割り込みによって発行するシグナル(one-shot)を設定します。
#define REX_DIO_IOCS_INTSIG_ENABLE	_IOW(REX_DIO_IOC_MAGIC, 6, T_IOCTL_DIO_INT_ENABLE)

// 割り込みシグナル解除
// ※REX_DIO_IOCS_INTSIG_ENABLEで設定したシグナルを解除します。
#define REX_DIO_IOCS_INTSIG_DISABLE	_IOW(REX_DIO_IOC_MAGIC, 7, T_IOCTL_DIO_INT_DISABLE)

// PORT 入出力方向設定
// ※MPIOSELレジスタに入出力方向を書き込みます
#define REX_DIO_IOCS_DIRECTION		_IOW(REX_DIO_IOC_MAGIC, 8, T_IOCTL_DIO_DIRECTION)

// PORT 入力反転設定
// ※MPIOINVレジスタに入力反転設定を書き込みます
#define REX_DIO_IOCS_INVERT			_IOW(REX_DIO_IOC_MAGIC, 9, T_IOCTL_DIO_INVERT)

// PORT TRI-STATE設定
// ※MPIO3TレジスタにTRI-STATE設定を書き込みます
#define REX_DIO_IOCS_TRISTATE		_IOW(REX_DIO_IOC_MAGIC, 10, T_IOCTL_DIO_TRISTATE)

#define REX_DIO_IOC_MAXNR			10

#endif /* _REXPX6X70D_DRV_H */

