/** 
 * @file rexpx6x70d_dio.h
 * @brief DIO入出力ライブラリ
 * @version 1.0.0
 * @date 2012-02-17
 * @par Copyright
 *  (C) 2012 RATOC Systems, Inc.
 * @par History
 *  - 2012.02.17 v1.0.0
 *    -# Initial Version.
 */
/*
 *  Copyright (C) 2012 RATOC Systems, Inc.
 *  Dual licensed under the BSD (BSD-LICENSE.txt)
 *  and GPL (GPL-LICENSE.txt) licenses.
 */

#ifndef __REXPX6X70D_DIO_H
#define __REXPX6X70D_DIO_H

/* ------------------------- */
/*	DEFINE^  ^   ^   ^   ^   */
/* ------------------------- */
/** 
 * @enum board_kind
 * @brief PEとPCI選択用の定義。DioOpenの引数にて使用する。
 */
typedef enum board_kind {
	eBOARD_KIND_PE = 0,													/**< REX-PExxを選択									*/
	eBOARD_KIND_PCI,													/**< REX-PCIxxを選択								*/
	eBOARD_KIND_MAX														/**< 最大値											*/
} board_kind_e;															/**< DioOpenの引数									*/

typedef void (*PAPIFUNC)(void);

/** 
 * @enum retcode_lib
 * @brief ライブラリのエラーコード
 */
typedef enum retcode_lib {
	eLIB_OK					= 0,										/**< success										*/
	eLIB_ERR_DIRECTION		= -1,										/**< direction error								*/
	eLIB_ERR_DRIVER			= -2,										/**< driver call error								*/
	eLIB_ERR_PARAM			= -3,										/**< parameter error								*/
	eLIB_ERR_INTERRUPT		= -4,										/**< already interrupt start						*/
	eLIB_ERR_TIMEOUT		= -5,										/**< interrupt timeout								*/
} retcode_lib_e;														/**<エラーコード									*/

/* ------------------------ */
/*	PUBLIC PROTTYPE	^   ^   */
/* ------------------------ */
int DioOpen(int BoardID, board_kind_e BoardKind);
int DioClose(int fd);
int SetDirection(int fd, unsigned char Direction);
int ReadPort(int fd, unsigned char *pReadData);
int WritePort(int fd, unsigned char WriteData);
int InterruptStart(int fd, unsigned char IntBit, unsigned char IntEdge, PAPIFUNC pCallBack);
int InterruptStop(int fd);
int InterruptStartSignal(int fd, unsigned char IntBit, unsigned char IntEdge, int nSignal);
int InterruptStopSignal(int fd, int nSignal);

#endif /* __REXPX6X70D_DIO_H */

